/***********************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 * Actuate Corporation - initial API and implementation
 ***********************************************************************/

package org.eclipse.birt.chart.reportitem.ui.dialogs;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.eclipse.birt.chart.reportitem.ui.i18n.Messages;
import org.eclipse.birt.report.designer.core.model.SessionHandleAdapter;
import org.eclipse.birt.report.designer.ui.views.attributes.IPropertyDescriptor;
import org.eclipse.birt.report.designer.ui.views.attributes.providers.ChoiceSetFactory;
import org.eclipse.birt.report.designer.ui.views.attributes.providers.PropertyProcessor;
import org.eclipse.birt.report.model.api.CommandStack;
import org.eclipse.birt.report.model.api.DataSetHandle;
import org.eclipse.birt.report.model.api.DataSetParameterHandle;
import org.eclipse.birt.report.model.api.DesignEngine;
import org.eclipse.birt.report.model.api.ParamBindingHandle;
import org.eclipse.birt.report.model.api.PropertyHandle;
import org.eclipse.birt.report.model.api.ReportItemHandle;
import org.eclipse.birt.report.model.api.StructureFactory;
import org.eclipse.birt.report.model.api.activity.SemanticException;
import org.eclipse.birt.report.model.api.elements.DesignChoiceConstants;
import org.eclipse.birt.report.model.api.elements.ReportDesignConstants;
import org.eclipse.birt.report.model.api.elements.structures.ParamBinding;
import org.eclipse.birt.report.model.api.metadata.IChoiceSet;
import org.eclipse.birt.report.model.api.metadata.PropertyValueException;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ICellModifier;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FormAttachment;
import org.eclipse.swt.layout.FormData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Item;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;

/**
 * ChartDataBindingPage
 */
public class ChartDataBindingPage extends Composite
{

	List input = null;

	/**
	 * The list kept Property & PropertyDescriptor pair.
	 */
	protected HashMap propertiesMap = new HashMap( 7 );

	/**
	 * The Binding properties table.
	 */
	Table table;

	/**
	 * The DataSet choose control.
	 */
	Combo combo;

	private ArrayList resultList;
	private ArrayList bindingParametersList;

	/**
	 * The static String value serves as key to <code>SetData(key,value)</code>
	 * method.
	 */
	public static final String Binding = "binding"; //$NON-NLS-1$

	/**
	 * The TableViewer of the table widget.
	 */
	private TableViewer tableViewer;

	/**
	 * The column list.
	 */
	static final String[] columnNames = {
			Messages.getString( "ChartDataBindingPage.Lbl.Parameter" ), Messages.getString( "ChartDataBindingPage.Lbl.DataType" ), Messages.getString( "ChartDataBindingPage.Lbl.Value" ) //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	};

	static final String NONE = Messages.getString( "ChartDataBindingPage.Lbl.None" ); //$NON-NLS-1$

	private transient boolean enableAutoCommit = true;

	private ExpressionDialogCellEditor expressionCellEditor;

	static IChoiceSet DataTypes = DesignEngine.getMetaDataDictionary( )
			.getChoiceSet( DesignChoiceConstants.CHOICE_COLUMN_DATA_TYPE );

	static final String DEFAULT_VALUE_LABEL = Messages.getString( "ChartDataBindingPage.Lbl.DefaultValue" ); //$NON-NLS-1$

	static final String DATA_SET_LABEL = Messages.getString( "ChartDataBindingPage.Lbl.DataSet" ); //$NON-NLS-1$

	/**
	 * @param parent
	 *            A widget which will be the parent of the new instance (cannot
	 *            be null)
	 * @param style
	 *            The style of widget to construct
	 */
	public ChartDataBindingPage( Composite parent, int style )
	{
		super( parent, style );
		buildUI( );
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.internal.ui.views.attributes.page.AttributePage#buildUI()
	 */
	protected void buildUI( )
	{
		// sets the layout
		FormLayout layout = new FormLayout( );
		layout.marginHeight = 7;
		layout.marginWidth = 7;
		layout.spacing = 5;
		setLayout( layout );

		FormData data;

		Label title = new Label( this, SWT.NONE );
		PropertyProcessor processor = new PropertyProcessor( ReportDesignConstants.REPORT_ITEM,
				ReportItemHandle.DATA_SET_PROP );
		title.setText( processor.getDisplayName( ) );

		combo = new Combo( this, SWT.READ_ONLY );
		combo.addSelectionListener( new SelectionAdapter( ) {

			public void widgetSelected( SelectionEvent e )
			{
				String value = combo.getText( );
				if ( value.equals( NONE ) )
					value = null;
				if ( canChangeDataSet( value ) )
				{
					try
					{
						startTrans( "" ); //$NON-NLS-1$
						DataSetHandle dataSet = null;
						if ( value != null )
						{
							dataSet = SessionHandleAdapter.getInstance( )
									.getReportDesignHandle( )
									.findDataSet( value );
						}
						getReportItemHandle( ).setDataSet( dataSet );
						getPropertyHandle( ).setStringValue( null );
						commit( );
					}
					catch ( SemanticException e1 )
					{
						rollback( );
						e1.printStackTrace( );
					}
					Object[] descriptors = propertiesMap.values( ).toArray( );
					for ( int i = 0; i < descriptors.length; i++ )
					{
						IPropertyDescriptor descriptor = (IPropertyDescriptor) descriptors[i];
						descriptor.setModelList( input );
					}
					refreshValues( );
				}
				else
				{
					combo.setText( value );
				}
			}
		} );
		data = new FormData( );
		data.left = new FormAttachment( title, 0, SWT.RIGHT );
		data.top = new FormAttachment( title, 0, SWT.CENTER );
		data.right = new FormAttachment( 50 );
		combo.setLayoutData( data );

		// create table and tableViewer
		table = new Table( this, SWT.SINGLE
				| SWT.BORDER
				| SWT.H_SCROLL
				| SWT.V_SCROLL
				| SWT.FULL_SELECTION );
		table.setLinesVisible( true );
		table.setHeaderVisible( true );
		for ( int i = 0; i < columnNames.length; i++ )
		{
			TableColumn column = new TableColumn( table, SWT.LEFT );
			column.setText( columnNames[i] );
			if ( i == 1 )
			{
				column.setWidth( 80 );
			}
			else
			{
				column.setWidth( 160 );
			}
		}
		// layout table
		data = new FormData( );
		data.top = new FormAttachment( combo, 0, SWT.BOTTOM );
		data.left = new FormAttachment( title, 0, SWT.LEFT );
		data.right = new FormAttachment( 100 );
		data.bottom = new FormAttachment( 100 );
		table.setLayoutData( data );
		createTableViewer( );
	}

	/**
	 * Creates the TableViewer and set all kinds of processors.
	 */
	private void createTableViewer( )
	{
		tableViewer = new TableViewer( table );
		tableViewer.setUseHashlookup( true );
		tableViewer.setColumnProperties( columnNames );
		expressionCellEditor = new ExpressionDialogCellEditor( table );
		tableViewer.setCellEditors( new CellEditor[]{
				null, null, expressionCellEditor
		} );
		tableViewer.setContentProvider( new BindingContentProvider( ) );
		tableViewer.setLabelProvider( new BindingLabelProvider( ) );
		tableViewer.setCellModifier( new BindingCellModifier( ) );
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.internal.ui.views.attributes.page.AttributePage#refreshValues(java.util.Set)
	 */
	protected void refreshValues( )
	{
		// Binding Not support multi-selection.
		if ( input.size( ) != 1 )
		{
			combo.setEnabled( false );
			combo.deselectAll( );
			table.removeAll( );
			table.setEnabled( false );
			return;
		}
		combo.setEnabled( true );
		table.setEnabled( true );

		String selectedDataSetName = combo.getText( );
		String[] oldList = combo.getItems( );
		String[] dataSets = ChoiceSetFactory.getDataSets( );
		String[] newList = new String[dataSets.length + 1];
		newList[0] = NONE;
		System.arraycopy( dataSets, 0, newList, 1, dataSets.length );
		if ( !Arrays.asList( oldList ).equals( Arrays.asList( newList ) ) )
		{
			combo.setItems( newList );
			combo.setText( selectedDataSetName );
		}
		String dataSetName = getDataSetName( );
		if ( !dataSetName.equals( selectedDataSetName ) )
		{
			combo.deselectAll( );
			combo.setText( dataSetName );
		}
		reconstructTable( );
		updateBindingData( );
	}

	ReportItemHandle getReportItemHandle( )
	{
		return (ReportItemHandle) input.get( 0 );
	}

	private String getDataSetName( )
	{
		if ( getReportItemHandle( ).getDataSet( ) == null )
		{
			return NONE;
		}
		String dataSetName = getReportItemHandle( ).getDataSet( )
				.getQualifiedName( );
		if ( dataSetName == null || "".equals( dataSetName ) ) //$NON-NLS-1$
		{
			dataSetName = NONE;
		}
		return dataSetName;
	}

	/**
	 * reconstruct the content of the table to show the last parameters in
	 * DataSet.
	 */
	private void reconstructTable( )
	{
		ReportItemHandle reportItemHandle = (ReportItemHandle) input.get( 0 );
		// DataSetHandle dataHandle = (DataSetHandle)
		// reportItemHandle.getDataSet();
		tableViewer.setInput( reportItemHandle );
		// tableViewer.refresh( );
		expressionCellEditor.setItemHandle( reportItemHandle );
	}

	// /**
	// * reconstruct the content of the table to show the last parameters in
	// DataSet.
	// */
	// private void constructTable()
	// {
	// ReportItemHandle reportItemHandle = (ReportItemHandle) input.get(0);
	// DataSetHandle dataHandle = (DataSetHandle) reportItemHandle.getDataSet();
	// tableViewer.setInput(dataHandle);
	// expressionCellEditor.setDataSetList(DEUtil.getDataSetList(reportItemHandle));
	// }

	/**
	 * Sets text of Value column
	 */
	private void updateBindingData( )
	{
		if ( input.size( ) != 1 )
			return;
		ReportItemHandle reportItemHandle = (ReportItemHandle) input.get( 0 );
		if ( reportItemHandle.getDataSet( ) == null )
			return;
		Iterator iterator = reportItemHandle.paramBindingsIterator( );
		while ( iterator != null && iterator.hasNext( ) )
		{
			ParamBindingHandle handle = (ParamBindingHandle) iterator.next( );
			String expression = handle.getExpression( );
			int rowIndex = this.bindingParametersList.indexOf( handle );
			if ( rowIndex != -1 && expression != null )
			{
				table.getItem( rowIndex ).setText( columnNames.length - 1,
						expression );
				Item item = table.getItem( rowIndex );
				if ( item.getData( Binding ) == null )
					item.setData( Binding, handle );
			}
		}
	}

	/**
	 * Creates a new ParamBinding Handle.
	 * 
	 * @return ParamBinding Handle.
	 * @throws SemanticException
	 */
	ParamBindingHandle createBindingHandle( String name )
			throws SemanticException
	{
		PropertyHandle propertyHandle = getPropertyHandle( );
		ParamBinding binding = StructureFactory.createParamBinding( );
		binding.setParamName( name );
		propertyHandle.addItem( binding );
		return (ParamBindingHandle) binding.getHandle( propertyHandle );
	}

	/**
	 * Gets the PropertyHandle of PARAM_BINDINGS_PROP property.
	 * 
	 * @return PropertyHandle
	 */
	PropertyHandle getPropertyHandle( )
	{
		ReportItemHandle handle = (ReportItemHandle) input.get( 0 );
		return handle.getPropertyHandle( ReportItemHandle.PARAM_BINDINGS_PROP );
	}

	private class BindingLabelProvider extends LabelProvider implements
			ITableLabelProvider
	{

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnImage(java.lang.Object,
		 *      int)
		 */
		public Image getColumnImage( Object element, int columnIndex )
		{
			return null;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnText(java.lang.Object,
		 *      int)
		 */
		public String getColumnText( Object element, int columnIndex )
		{
			String text = ""; //$NON-NLS-1$
			DataSetParameterHandle parameter = (DataSetParameterHandle) ( (Object[]) element )[0];
			ParamBindingHandle bindingParameter = (ParamBindingHandle) ( (Object[]) element )[1];
			switch ( columnIndex )
			{
				case 0 :
					if ( parameter.getName( ) != null )
					{
						text = parameter.getName( );
					}
					break;
				case 1 :
					if ( parameter.getDataType( ) != null )
					{
						text = ChoiceSetFactory.getDisplayNameFromChoiceSet( parameter.getDataType( ),
								DataTypes );
					}
					break;
				case 2 :
					if ( bindingParameter != null
							&& bindingParameter.getExpression( ) != null )
					{
						text = bindingParameter.getExpression( );
					}
					else if ( parameter.getDefaultValue( ) != null )
					{
						text = parameter.getDefaultValue( ) + " " //$NON-NLS-1$
								+ DEFAULT_VALUE_LABEL;
					}
					break;
			}
			return text;
		}
	}

	private class BindingCellModifier implements ICellModifier
	{

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.ICellModifier#canModify(java.lang.Object,
		 *      java.lang.String)
		 */
		public boolean canModify( Object element, String property )
		{
			return property.equals( columnNames[2] );
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.ICellModifier#getValue(java.lang.Object,
		 *      java.lang.String)
		 */
		public Object getValue( Object element, String property )
		{
			ParamBindingHandle bindingParameter = (ParamBindingHandle) ( (Object[]) element )[1];
			if ( bindingParameter != null )
			{
				return bindingParameter.getExpression( );
			}
			return null;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.ICellModifier#modify(java.lang.Object,
		 *      java.lang.String, java.lang.Object)
		 */
		public void modify( Object element, String property, Object value )
		{
			Object model = element;
			if ( element instanceof Item )
			{
				model = ( (Item) element ).getData( );
			}
			int index = resultList.indexOf( model );
			// if (index != -1)
			/*
			 * { Object[] pair = (Object[])model; ParamBindingHandle
			 * bindingHandle = (ParamBindingHandle)pair[1]; if ( bindingHandle !=
			 * null) { bindingHandle.setExpression( (String)value );
			 * reconstructTable( ); } }
			 */
			// remove the ParamBindingHandle
			if ( index != -1 )
			{
				Object[] pair = (Object[]) model;
				// DataSetParameterHandle dataHandle =
				// (DataSetParameterHandle)pair[0];
				ParamBindingHandle bindingHandle = (ParamBindingHandle) pair[1];
				// if value is reset, remove the ParamBindingHandle
				if ( ( value == null || "".equals( value ) ) && bindingHandle != null ) //$NON-NLS-1$
				{

					try
					{
						getPropertyHandle( ).removeItem( bindingHandle.getStructure( ) );
					}
					catch ( PropertyValueException e )
					{
						e.printStackTrace( );
						return;
					}
				}
				// if no value yet, create a new ParamBindingHandle
				startTrans( Messages.getString( "ChartDataBindingPage.Lbl.SaveParameterBinding" ) ); //$NON-NLS-1$
				if ( bindingHandle == null )
				{
					try
					{
						bindingHandle = createBindingHandle( table.getItem( index )
								.getText( 0 ) );
					}
					catch ( SemanticException e )
					{
						e.printStackTrace( );
						rollback( );
						return;
					}
				}

				bindingHandle.setExpression( (String) value );
				commit( );
				reconstructTable( );
			}
		}
	}

	private class BindingContentProvider implements IStructuredContentProvider
	{

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.IStructuredContentProvider#getElements(java.lang.Object)
		 */
		public Object[] getElements( Object inputElement )
		{
			if ( inputElement == null )
			{
				return new Object[0];
			}
			ReportItemHandle inputHandle = (ReportItemHandle) inputElement;
			DataSetHandle dataHandle = inputHandle.getDataSet( );
			if ( dataHandle == null )
			{
				return new Object[0];
			}
			bindingParametersList = new ArrayList( );
			List bindingParametersNameList = new ArrayList( );
			resultList = new ArrayList( );
			for ( Iterator iterator = inputHandle.paramBindingsIterator( ); iterator.hasNext( ); )
			{
				ParamBindingHandle handle = (ParamBindingHandle) iterator.next( );
				bindingParametersList.add( handle );
				bindingParametersNameList.add( handle.getParamName( ) );
			}
			for ( Iterator iterator = dataHandle.parametersIterator( ); iterator.hasNext( ); )
			{
				DataSetParameterHandle handle = (DataSetParameterHandle) iterator.next( );
				Object[] result = new Object[]{
						handle, null
				};
				int index = bindingParametersNameList.indexOf( handle.getName( ) );
				if ( index != -1 )
				{
					result[1] = bindingParametersList.get( index );
				}
				resultList.add( result );
			}
			return resultList.toArray( );
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.IContentProvider#dispose()
		 */
		public void dispose( )
		{
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.viewers.IContentProvider#inputChanged(org.eclipse.jface.viewers.Viewer,
		 *      java.lang.Object, java.lang.Object)
		 */
		public void inputChanged( Viewer viewer, Object oldInput,
				Object newInput )
		{
		}
	}

	/**
	 * Gets the DE CommandStack instance
	 * 
	 * @return CommandStack instance
	 */
	private CommandStack getActionStack( )
	{
		return SessionHandleAdapter.getInstance( ).getCommandStack( );
	}

	void startTrans( String name )
	{
		if ( isEnableAutoCommit( ) )
		{
			getActionStack( ).startTrans( name );
		}
	}

	void commit( )
	{
		if ( isEnableAutoCommit( ) )
		{
			getActionStack( ).commit( );
		}
	}

	void rollback( )
	{
		if ( isEnableAutoCommit( ) )
		{
			getActionStack( ).rollback( );
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.ui.attributes.page.TabPage#setInput(java.util.List)
	 */
	public void setInput( List elements )
	{
		input = elements;
		reconstructTable( );

		Object[] descriptors = propertiesMap.values( ).toArray( );
		for ( int i = 0; i < descriptors.length; i++ )
		{
			IPropertyDescriptor descriptor = (IPropertyDescriptor) descriptors[i];
			descriptor.setModelList( elements );
		}
		refreshValues( );
	}

	/**
	 * @return Returns the enableAutoCommit.
	 */
	public boolean isEnableAutoCommit( )
	{
		return enableAutoCommit;
	}

	/**
	 * @param enableAutoCommit
	 *            The enableAutoCommit to set.
	 */
	public void setEnableAutoCommit( boolean enableAutoCommit )
	{
		this.enableAutoCommit = enableAutoCommit;
	}

	boolean canChangeDataSet( String newName )
	{
		if ( isEnableAutoCommit( ) )
		{
			PropertyProcessor processor = new PropertyProcessor( ReportDesignConstants.TABLE_ITEM,
					ReportItemHandle.DATA_SET_PROP );
			String currentDataSetName = processor.getStringValue( input );
			if ( currentDataSetName == null
					|| "".equals( currentDataSetName ) || NONE.equals( currentDataSetName ) ) //$NON-NLS-1$
			{
				return true;
			}
			else if ( !currentDataSetName.equals( newName ) )
			{
				return MessageDialog.openQuestion( null,
						Messages.getString( "ChartDataBindingPage.Lbl.SaveDataSet" ), Messages.getString( "ChartDataBindingPage.Msg.AskSaveDataSet" ) ); //$NON-NLS-1$ //$NON-NLS-2$
			}

			return false;
		}
		return true;
	}
}