/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.core.archive.compound;

import java.io.EOFException;
import java.io.IOException;
import java.util.HashMap;

import org.eclipse.birt.core.util.IOUtil;

/**
 * The AllocationTable defines the FAT table in the archive file.
 * 
 * The FAT table a integer array, each element represent a block in the stream.
 * The block id is the array index. The value of the element is the id of the
 * next block followed by the block index. It constructs a link list which saves
 * the blocks in a stream.
 * <p>
 * There are multiple links in the FAT table.
 * <li>The first link stream is the free blocks which starts from block 0.</li>
 * <li>The second link stream is the FAT table itself which starts from block
 * 1.</li>
 * <li>The third link stream is the entry table which starts from block 2.</li>
 * <li>The other streams id are saved in the entry table.</li>
 */
class AllocationTable
{
	/**
	 * the increase step of the size of the link list.
	 */
	final static int BLOCK_COUNT_INCREASE = 32;
	/**
	 * the first block in the FAT table
	 */
	final static int FIRST_FAT_BLOCK = 1;
	/**
	 * the first block in the free list.
	 */
	final static int FIRST_FREE_BLOCK = 0;
	/**
	 * it means that the block list is ended.
	 */
	final static int LAST_BLOCK = -1;
	/**
	 * it means that the block continues in the disk
	 */
	final static int CONTINUE_BLOCK = 0;

	/**
	 * the file of the FAT.
	 */
	private ArchiveFile af;

	/**
	 * the list saves the blocks of the FAT.
	 */
	private int[] fatBlocks;
	/**
	 * total blocks in the FAT.
	 */
	private int totalFATBlocks;

	private HashMap entries;

	private AllocationTable( ArchiveFile af )
	{
		this.af = af;
		fatBlocks = new int[BLOCK_COUNT_INCREASE];
		fatBlocks[0] = FIRST_FAT_BLOCK;
		totalFATBlocks = 1;
		entries = new HashMap( );
	}

	static AllocationTable createTable( ArchiveFile af ) throws IOException
	{
		AllocationTable table = new AllocationTable( af );
		table.initFATBlock( FIRST_FAT_BLOCK );
		return table;
	}

	static AllocationTable loadTable( ArchiveFile af ) throws IOException
	{
		AllocationTable table = new AllocationTable( af );
		table.refresh( );
		return table;
	}

	/**
	 * write the whole FAT table into the file.
	 * 
	 * @throws IOException
	 */
	void flush( ) throws IOException
	{
		// flush all the fat tables into the disk
		// TODO: flush all blocks which owned by the fat table
	}

	/**
	 * reload the FAT table from the file.
	 * 
	 * @throws IOException
	 */
	void refresh( ) throws IOException
	{
		// reloads all the blocks in the fat table from the disk.
		// TODO: rerefesh all the blocks which owned by the FAT table
		int lastBlockId = fatBlocks[totalFATBlocks - 1];
		int blockId = readFATInt( lastBlockId * 4 );
		while ( blockId != -1 )
		{
			ensureFATBlocks( totalFATBlocks + 1 );
			fatBlocks[totalFATBlocks] = blockId;
			totalFATBlocks++;
			blockId = readFATInt( blockId * 4 );
		}
	}

	/**
	 * allocate a block from the free list. If the free list is empty, create a
	 * new block at the end of file.
	 * 
	 * @return the block id.
	 * @throws IOException
	 */
	private int getFreeBlock( ) throws IOException
	{
		// get the free block
		int freeBlockId = readFATInt( FIRST_FREE_BLOCK );
		if ( freeBlockId != -1 )
		{
			int nextFreeBlockId = readFATInt( freeBlockId * 4 );
			writeFATInt( FIRST_FREE_BLOCK, nextFreeBlockId );
			return freeBlockId;
		}
		// create a new block from the streams
		return af.allocateBlock( );
	}

	/**
	 * create a entry in the FAT table. the entry will start from the first free
	 * block.
	 * 
	 * @return the created entry
	 * @throws IOException
	 */
	Entry createEntry( ) throws IOException
	{
		int blockId = getFreeBlock( );
		writeFATInt( blockId * 4, -1 );
		Entry entry = new Entry( blockId );
		entries.put( new Integer( blockId ), entry );
		return entry;
	}

	/**
	 * load the entry from the FAT table.
	 * 
	 * @param blockId
	 *            the first block of the entry.
	 * @return the entry loaded.
	 * @throws IOException
	 */
	Entry loadEntry( int blockId ) throws IOException
	{
		Entry entry = (Entry) entries.get( new Integer( blockId ) );
		if ( entry == null )
		{
			entry = new Entry( blockId );
			entry.refresh( );
			entries.put( new Integer( blockId ), entry );
		}
		return entry;
	}

	/**
	 * remove the entry from the FAT table.
	 * 
	 * @param entry
	 *            the entry to be removed.
	 * @throws IOException
	 */
	void removeEntry( Entry entry ) throws IOException
	{
		int lastBlockId = entry.getLastBlock( );
		int firstBlockId = entry.getFirstBlock( );
		int freeBlockId = readFATInt( FIRST_FREE_BLOCK );
		writeFATInt( lastBlockId * 4, freeBlockId );
		writeFATInt( FIRST_FREE_BLOCK, firstBlockId );
		entries.remove( new Integer( firstBlockId ) );
	}

	private int readFATInt( long offset ) throws IOException
	{
		if ( offset > totalFATBlocks * Block.BLOCK_SIZE )
		{
			throw new EOFException( );
		}
		int blockId = (int) ( offset / Block.BLOCK_SIZE );
		int off = (int) ( offset % Block.BLOCK_SIZE );
		int phyBlockId = getFATBlock( blockId );
		byte[] b = new byte[4];
		af.read( phyBlockId, off, b, 0, 4 );
		return IOUtil.bytesToInteger( b );
	}

	private void writeFATInt( long offset, int block ) throws IOException
	{
		int blockId = (int) ( offset / Block.BLOCK_SIZE );
		int off = (int) ( offset % Block.BLOCK_SIZE );
		if ( blockId >= totalFATBlocks )
		{
			int newTotalBlocks = blockId + 1;
			ensureFATBlocks( newTotalBlocks );
			int lastFATBlock = fatBlocks[totalFATBlocks - 1];
			for ( int i = totalFATBlocks; i < newTotalBlocks; i++ )
			{
				int freeBlock = getFreeBlock( );
				writeFATInt( lastFATBlock * 4, freeBlock );
				fatBlocks[totalFATBlocks] = freeBlock;
				lastFATBlock = freeBlock;
				totalFATBlocks++;
			}
			writeFATInt( lastFATBlock * 4, LAST_BLOCK );
		}

		int phyBlockId = getFATBlock( blockId );
		byte[] b = new byte[4];
		IOUtil.integerToBytes( block, b );
		af.write( phyBlockId, off, b, 0, 4 );
	}

	private void ensureFATBlocks( int size )
	{
		// ensure the buffer is larger enough.
		if ( fatBlocks.length < size )
		{
			int length = ( size / BLOCK_COUNT_INCREASE + 1 )
					* BLOCK_COUNT_INCREASE;
			int[] blocks = new int[length];
			System.arraycopy( fatBlocks, 0, blocks, 0, totalFATBlocks );
			fatBlocks = blocks;
		}
	}

	private int getFATBlock( int blockId ) throws IOException
	{
		assert blockId < totalFATBlocks;
		return fatBlocks[blockId];
	}

	/**
	 * initialize the block id as FAT block. In the FAT block, all index should
	 * be set to -1.
	 * 
	 * @param blockId
	 *            the id of the block to be initalize.
	 * @throws IOException
	 */
	private void initFATBlock( int blockId ) throws IOException
	{
		byte[] b = new byte[Block.BLOCK_SIZE];
		for ( int i = 0; i < Block.BLOCK_SIZE; i++ )
		{
			b[i] = -1;
		}
		af.write( blockId, 0, b, 0, Block.BLOCK_SIZE );
	}

	/**
	 * the entry defined in the FAT stream. It defines the blocks in a stream.
	 * 
	 */
	class Entry
	{

		private int totalBlocks;
		private int[] blockIds;

		private Entry( int blockId ) throws IOException
		{
			blockIds = new int[BLOCK_COUNT_INCREASE];
			blockIds[0] = blockId;
			totalBlocks = 1;
		}

		void refresh( ) throws IOException
		{
			int lastBlockId = getLastBlock( );
			int blockId = readFATInt( lastBlockId * 4 );
			while ( blockId != -1 )
			{
				ensureBlocks( totalBlocks + 1 );
				blockIds[totalBlocks] = blockId;
				totalBlocks++;
				blockId = readFATInt( blockId * 4 );
			}
		}

		private void ensureBlocks( int size )
		{
			// ensure the buffer is larger enough.
			if ( blockIds == null || blockIds.length < size )
			{
				size = ( size / BLOCK_COUNT_INCREASE ) * BLOCK_COUNT_INCREASE
						+ BLOCK_COUNT_INCREASE;
				int[] blocks = new int[size];
				if ( blockIds != null )
				{
					System.arraycopy( blockIds, 0, blocks, 0, totalBlocks );
				}
				blockIds = blocks;
			}
		}

		int getTotalBlocks( ) throws IOException
		{
			return totalBlocks;
		}

		int getFirstBlock( )
		{
			return blockIds[0];
		}

		int getLastBlock( )
		{
			return blockIds[totalBlocks - 1];
		}

		int getBlock( int index )
		{
			if ( index < totalBlocks )
			{
				return blockIds[index];
			}
			return -1;
		}

		void allocBlocks( int blockCount ) throws IOException
		{
			ensureBlocks( totalBlocks + blockCount );
			int lastBlockId = getLastBlock( );
			for ( int i = 0; i < blockCount; i++ )
			{
				int blockId = getFreeBlock( );
				blockIds[totalBlocks + i] = blockId;
				writeFATInt( lastBlockId * 4, blockId );
				lastBlockId = blockId;
			}
			writeFATInt( lastBlockId * 4, -1 );
			totalBlocks += blockCount;
		}
	}
}
