/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.core.archive.compound;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;

import org.eclipse.birt.core.util.IOUtil;

/**
 * A table contains all the stream item blocks of a compound file system.
 */
class EntryTable
{

	/**
	 * the first block in the entry table
	 */
	final static int FIRST_ENTRY_BLOCK = 2;

	protected int totalSlots;
	protected HashMap entries;
	protected AllocationTable.Entry index;
	protected ArchiveFile af;

	private EntryTable( ArchiveFile af ) throws IOException
	{
		this.af = af;
		totalSlots = 0;
		entries = new HashMap( );
		index = af.allocTbl.loadEntry( FIRST_ENTRY_BLOCK );
	}

	static EntryTable loadTable( ArchiveFile af ) throws IOException
	{
		EntryTable table = new EntryTable( af );
		table.loadEntries( );
		return table;
	}

	static EntryTable createTable( ArchiveFile af ) throws IOException
	{
		EntryTable table = new EntryTable( af );
		table.initEntries( );
		return table;
	}

	private void loadEntries( ) throws IOException
	{
		// refresh the
		totalSlots = 0;
		do
		{
			Entry entry = new Entry( totalSlots );
			entry.read( );
			int usedSlots = entry.getUsedSlots( );
			if ( usedSlots == 0 )
			{
				totalSlots += 1;
				continue;
			}
			if ( usedSlots == -1 )
			{
				break;
			}
			entries.put( entry.getName( ), entry );
			totalSlots += usedSlots;
		} while ( true );
	}

	private void initEntries( ) throws IOException
	{
		totalSlots = 0;
		setUsedSlots( 0, -1 );
	}

	void removeEntry( Entry entry ) throws IOException
	{
		entries.remove( entry.getName( ) );
		int slotId = entry.getSlotID( );
		int usedSlots = entry.getUsedSlots( );
		for ( int i = 0; i < usedSlots; i++ )
		{
			setUsedSlots( slotId + i, 0 );
		}
	}

	/**
	 * create a stream item
	 * 
	 * @param name
	 *            stream name
	 * @return the stream item created
	 * @throws IOException
	 */

	Entry createEntry( String name ) throws IOException
	{
		Entry entry = (Entry) entries.get( name );
		if ( entry != null )
		{
			entry.setLength( 0 );
			return entry;
		}

		ensureSlots( totalSlots + 1 );
		entry = new Entry( totalSlots, name );
		entry.write( );
		totalSlots += entry.getUsedSlots( );
		setUsedSlots( totalSlots, -1 );

		entries.put( name, entry );

		return entry;
	}

	Entry findEntry( String name ) throws IOException
	{
		return (Entry) entries.get( name );
	}

	Collection listEntries( )
	{
		return entries.values( );
	}

	private void ensureSlots( int slotSize ) throws IOException
	{
		// the last slot is used to indicate the EOF table
		long offset = (long) ( slotSize + 1 ) * ENTRY_SLOT_SIZE;
		int blockCount = (int) ( ( offset + Block.BLOCK_SIZE - 1 ) / Block.BLOCK_SIZE );
		int totalBlock = index.getTotalBlocks( );
		if ( blockCount > totalBlock )
		{
			int increaseCount = blockCount - totalBlock + 1;
			index.allocBlocks( increaseCount );
		}
	}

	private void setUsedSlots( int slotId, int usedSlots ) throws IOException
	{
		ensureSlots( slotId + 1 );
		long offset = slotId * ENTRY_SLOT_SIZE + ENTRY_USED_SLOT_OFFSET;
		int blkId = (int) ( offset / Block.BLOCK_SIZE );
		int blkOff = (int) ( offset % Block.BLOCK_SIZE );
		int phyBlk = index.getBlock( blkId );
		byte[] b = new byte[4];
		IOUtil.integerToBytes( usedSlots, b );
		af.write( phyBlk, blkOff, b, 0, 4 );
	}

	static final int ENTRY_SLOT_SIZE = 128;
	static final int ENTRY_USED_SLOT_OFFSET = 0;
	static final int ENTRY_LENGTH_OFFSET = 4;
	static final int ENTRY_BLOCK_OFFSET = 12;
	static final int ENTRY_NAME_OFFSET = 16;

	void readSlot( int slotId, byte[] b, int off ) throws IOException
	{
		long offset = (long) slotId * ENTRY_SLOT_SIZE;
		int blkId = (int) ( offset / Block.BLOCK_SIZE );
		int blkOff = (int) ( offset % Block.BLOCK_SIZE );
		int phyBlk = index.getBlock( blkId );
		af.read( phyBlk, blkOff, b, off, ENTRY_SLOT_SIZE );
	}

	void writeSlot( int slotId, byte[] b, int off ) throws IOException
	{
		long offset = (long) slotId * ENTRY_SLOT_SIZE;
		int blkId = (int) ( offset / Block.BLOCK_SIZE );
		int blkOff = (int) ( offset % Block.BLOCK_SIZE );
		int phyBlk = index.getBlock( blkId );
		int size = b.length - off;
		if ( size > ENTRY_SLOT_SIZE )
		{
			size = ENTRY_SLOT_SIZE;
		}
		af.write( phyBlk, blkOff, b, off, size );
	}

	/**
	 * name entry in the entry table. Each entry represents a stream in the
	 * file.
	 * 
	 */
	class Entry
	{

		/**
		 * the slot id of the entry, it is used as the stream id.
		 */
		private int slotId;
		/**
		 * the slots used by this entry. For stream with long name, it may used
		 * mutiple slots.
		 */
		private int usedSlots;
		/**
		 * the length of the stream.
		 */
		private long length;
		/**
		 * the first block of this stream.
		 */
		private int firstBlock;
		/**
		 * the name of the stream.
		 */
		private String name;

		private Entry( int slotId )
		{
			this.slotId = slotId;
			this.name = null;
			this.usedSlots = 0;
			this.firstBlock = -1;
			this.length = 0;
		}

		private Entry( int slotId, String name )
		{
			this.slotId = slotId;
			this.name = name;
			this.usedSlots = 0;
			this.firstBlock = -1;
			this.length = 0;
		}

		void read( ) throws IOException
		{
			byte[] b = new byte[ENTRY_SLOT_SIZE];
			readSlot( slotId, b, 0 );
			usedSlots = IOUtil.bytesToInteger( b );
			if ( usedSlots <= 0 )
			{
				return;
			}
			if ( usedSlots > 1 )
			{
				byte[] nb = new byte[ENTRY_SLOT_SIZE * usedSlots];
				System.arraycopy( b, 0, nb, 0, ENTRY_SLOT_SIZE );
				for ( int i = 1; i < usedSlots; i++ )
				{
					readSlot( slotId + i, nb, i * ENTRY_SLOT_SIZE );
				}
				b = nb;
			}
			DataInputStream is = new DataInputStream( new ByteArrayInputStream(
					b ) );
			is.skipBytes( 4 ); // skip the used slots
			length = is.readLong( );
			firstBlock = is.readInt( );
			name = is.readUTF( );
		}

		void write( ) throws IOException
		{
			ByteArrayOutputStream buff = new ByteArrayOutputStream(
					ENTRY_SLOT_SIZE );
			DataOutputStream os = new DataOutputStream( buff );
			os.writeInt( 1 );
			os.writeLong( length );
			os.writeInt( firstBlock );
			os.writeUTF( name );
			byte[] b = buff.toByteArray( );
			usedSlots = ( b.length + ENTRY_SLOT_SIZE - 1 ) / ENTRY_SLOT_SIZE;
			if ( usedSlots != 1 )
			{
				IOUtil.integerToBytes( usedSlots, b );
			}
			for ( int i = 0; i < usedSlots; i++ )
			{
				writeSlot( slotId + i, b, i * ENTRY_SLOT_SIZE );
			}
		}

		int getSlotID( )
		{
			return slotId;
		}

		int getUsedSlots( )
		{
			return usedSlots;
		}

		String getName( )
		{
			return name;
		}

		long getLength( )
		{
			return length;
		}

		void setLength( long length ) throws IOException
		{
			long offset = slotId * ENTRY_SLOT_SIZE + ENTRY_LENGTH_OFFSET;
			int blkId = (int) ( offset / Block.BLOCK_SIZE );
			int blkOff = (int) ( offset % Block.BLOCK_SIZE );
			int phyBlk = index.getBlock( blkId );
			byte[] b = new byte[8];
			IOUtil.longToBytes( length, b );
			af.write( phyBlk, blkOff, b, 0, 8 );
			this.length = length;
		}

		int getBlock( )
		{
			return firstBlock;
		}

		void setBlock( int blockId ) throws IOException
		{
			long offset = slotId * ENTRY_SLOT_SIZE + ENTRY_BLOCK_OFFSET;
			int blkId = (int) ( offset / Block.BLOCK_SIZE );
			int blkOff = (int) ( offset % Block.BLOCK_SIZE );
			int phyBlk = index.getBlock( blkId );
			byte[] b = new byte[4];
			IOUtil.integerToBytes( blockId, b );
			af.write( phyBlk, blkOff, b, 0, 4 );
			this.firstBlock = blockId;
		}
	}
}
