/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/
package org.eclipse.birt.data.engine.executor.dscache;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.birt.core.data.DataType;
import org.eclipse.birt.core.data.ExpressionUtil;
import org.eclipse.birt.core.util.IOUtil;
import org.eclipse.birt.data.engine.api.querydefn.ComputedColumn;
import org.eclipse.birt.data.engine.api.querydefn.ScriptExpression;
import org.eclipse.birt.data.engine.core.DataException;
import org.eclipse.birt.data.engine.executor.DataSetCacheManager;
import org.eclipse.birt.data.engine.executor.ResultClass;
import org.eclipse.birt.data.engine.executor.ResultFieldMetadata;
import org.eclipse.birt.data.engine.executor.ResultObject;
import org.eclipse.birt.data.engine.executor.cache.ResultObjectUtil;
import org.eclipse.birt.data.engine.i18n.ResourceConstants;
import org.eclipse.birt.data.engine.impl.DataEngineSession;
import org.eclipse.birt.data.engine.odaconsumer.ResultSet;
import org.eclipse.birt.data.engine.odi.ICustomDataSet;
import org.eclipse.birt.data.engine.odi.IDataSetPopulator;
import org.eclipse.birt.data.engine.odi.IResultClass;
import org.eclipse.birt.data.engine.odi.IResultObject;

/**
 * Any data set result is supported to cache data into a temp file. In the first
 * time, the data can be retrieved from ODA driver. and then be stored into
 * cache file. In the second time or later, the data can be directly retrieved
 * from cache file rather than ODA driver to save time.
 */
public class DataSetResultCache
{
	//----------save time--------------
	
	// oda result set and metadata
	private IResultClass rsMeta;
	
	// oda data set or custom data set
	private ResultSet odaDataSet;
	private ICustomDataSet customDataSet;
	private IDataSetPopulator populator;
	
	// cache requirement
	private int cacheCount;
		
	// save util instance
	private SaveUtil saveUtil;
	private boolean isSaved;
	
	//----------load time--------------
	private LoadUtil loadUtil;
	
	// share info
	private boolean isLoad;
	
	// for computed column
	private List addedTempComputedColumn;
	private int realColumnCount;
	private DataEngineSession session;
	
	/**
	 * @param odaDataSet
	 * @param resultSet
	 */
	public DataSetResultCache( ResultSet odaDataSet, IResultClass rsMeta, DataEngineSession session )
	{
		assert odaDataSet != null;
		assert rsMeta != null;

		this.odaDataSet = odaDataSet;
		
		this.init( rsMeta, session );
	}
	
	/**
	 * @param customDataSet
	 * @param rsMeta
	 */
	public DataSetResultCache( IDataSetPopulator populator, IResultClass rsMeta, DataEngineSession session )
	{
		assert populator!=null;
		assert rsMeta != null;
		
		this.populator = populator;
		
		this.init( rsMeta, session );
	}
	
	/**
	 * @param customDataSet
	 * @param rsMeta
	 */
	public DataSetResultCache( ICustomDataSet customDataSet, IResultClass rsMeta, DataEngineSession session )
	{
		assert customDataSet!=null;
		assert rsMeta != null;
		
		this.customDataSet = customDataSet;
		this.init( rsMeta, session );
	}
	
	/**
	 * Initialize
	 */
	private void init( IResultClass rsMeta, DataEngineSession session )
	{
		this.rsMeta = rsMeta;
		this.isLoad = false;
		this.isSaved = false;
		this.session = session;
		this.cacheCount = getCacheCount( );
	}
	
	/**
	 * when loaded data, this constuction is used
	 */
	public DataSetResultCache(  DataEngineSession session )
	{
		this.isLoad = true;
		this.session = session;
	}
	
	/**
	 * set new temp computed columns
	 * @param addedTempComputedColumn
	 */
	public void setTempComputedColumn( List addedTempComputedColumn )
	{
		this.addedTempComputedColumn = addedTempComputedColumn;
	}
	
	/**
	 * @return next data
	 * @throws DataException 
	 */
	public IResultObject fetch( ) throws DataException
	{
		if ( isLoad == false )
		{
			if ( cacheCount <= 0 )
			{
				if ( odaDataSet != null )
					return odaDataSet.fetch( );
				else if ( customDataSet != null )
					return customDataSet.fetch( );
				else
					return populator.next( );
			}
			else
			{
				if ( isSaved == false )
				{
					cacheDataSet( );
					isSaved = true;
				}
				return loadObject( );
			}
		}		
		else
		{
			return loadObject( );
		}
	}
	
	/**
	 * @throws DataException
	 */
	private void cacheDataSet( ) throws DataException
	{
		int index = 0;
		IResultObject resultObject = null;

		this.saveInit( );
		
		do
		{
			if ( odaDataSet != null )
				resultObject = odaDataSet.fetch( );
			else if ( customDataSet != null )
				resultObject = customDataSet.fetch( );
			else
				resultObject = this.populator.next( );
			
			if ( resultObject != null )
			{
				saveUtil.saveObject( resultObject );
				index++;
				if ( index >= cacheCount )
				{
					break;
				}
			}
		} while ( resultObject != null );
		
		this.saveClose( );
	}
	
	/**
	 * @return result class
	 * @throws DataException 
	 */
	public IResultClass getResultClass( ) throws DataException
	{
		if ( isLoad == true && rsMeta == null )
		{
			rsMeta = loadResultClass( );
			if ( addedTempComputedColumn != null
					&& addedTempComputedColumn.size( ) > 0 )
				processResultClass( );
		}

		return rsMeta;
	}
	
	/**
	 * Remove old temp computed column metadatas from cache file and add new metadatas.
	 * @throws DataException
	 */
	private void processResultClass( ) throws DataException
	{
		List metadataList = new ArrayList( );
		this.realColumnCount = 0;
		
		ResultFieldMetadata metadata = null;
		int i = 0;
		for ( i = 0; i < rsMeta.getFieldCount( ); i++ )
		{
			if ( !isTempComputedColumn( rsMeta.getFieldName( i + 1 ) ) )
			{
				metadata = new ResultFieldMetadata( 0,
						rsMeta.getFieldName( i + 1 ),
						rsMeta.getFieldLabel( i + 1 ),
						rsMeta.getFieldValueClass( i + 1 ),
						rsMeta.getFieldNativeTypeName( i + 1 ),
						rsMeta.isCustomField( i + 1 ) );
				metadata.setAlias( rsMeta.getFieldAlias( i + 1 ) );
				metadataList.add( metadata );
				realColumnCount++;
			}
		}
		
		ComputedColumn tempComputedColumn = null;
		for ( i = 0; i < addedTempComputedColumn.size( ); i++ )
		{
			tempComputedColumn = (ComputedColumn) ( addedTempComputedColumn.get( i ) );
			metadata = new ResultFieldMetadata( 0,
					tempComputedColumn.getName( ),
					null,
					DataType.getClass( tempComputedColumn.getDataType( ) ),
					null,
					true );
			metadataList.add( metadata );
		}
		
		rsMeta = new ResultClass( metadataList );
	}
	
	/**
	 * Return whether a clumn is a temp computed column.
	 * @param name
	 * @return
	 */
	private boolean isTempComputedColumn(String name)
	{
		return ( name.matches( "\\Q_{$TEMP_GROUP_\\E\\d*\\Q$}_\\E" )
				|| name.matches( "\\Q_{$TEMP_SORT_\\E\\d*\\Q$}_\\E" )
				|| name.matches( "\\Q_{$TEMP_FILTER_\\E\\d*\\Q$}_\\E" ));
	}
	
	/**
	 * @throws DataException 
	 */
	public void close( ) throws DataException
	{
		// when in save status, close might be done automatically
		if ( loadUtil != null )
		{
			loadUtil.close( );
			loadUtil = null;
		}
	}
	
	/**
	 * Init save util
	 */
	private void saveInit( )
	{
		saveUtil = new SaveUtil( new File( getSaveFolder( ) + "/data.data" ),
				new File( getSaveFolder( ) + "/metaData.data" ),
				this.rsMeta );
	};
	
	/**
	 * @throws DataException
	 */
	private void saveClose( ) throws DataException
	{
		if ( saveUtil != null )
		{
			saveUtil.close( );
			saveUtil = null;
		}
	}
	
	/**
	 * @return cached object
	 * @throws DataException
	 */
	private IResultObject loadObject( ) throws DataException
	{
		if ( loadUtil == null )
			loadUtil = new LoadUtil( new File( getLoadFolder( ) + "/data.data" ),
					new File( getLoadFolder( ) + "/metaData.data" ) );

		IResultObject cacheObject = loadUtil.loadObject( );

		if ( cacheObject == null )
		{
			return cacheObject;
		}

		if ( addedTempComputedColumn != null
				&& addedTempComputedColumn.size( ) > 0 )
		{
			ResultObject resultObject = new ResultObject( getResultClass( ),
					getAllObjects( cacheObject ) );

			return resultObject;
		}
		else
		{
			return cacheObject;
		}
	}
	
	/**
	 * get all new field objects from a cacheObject
	 * @param cacheObject
	 * @return
	 * @throws DataException
	 */
	private Object[] getAllObjects( IResultObject cacheObject )
			throws DataException
	{
		Object[] objects = new Object[realColumnCount
				+ addedTempComputedColumn.size( )];
		for ( int i = 0; i < realColumnCount; i++ )
		{
			objects[i] = cacheObject.getFieldValue( i + 1 );
		}
		return objects;
	}
	
	/**
	 * @return IResultClass
	 * @throws DataException
	 */
	private IResultClass loadResultClass( ) throws DataException
	{
		if ( loadUtil == null )
			loadUtil = new LoadUtil( new File( getLoadFolder( ) + "/data.data" ),
					new File( getLoadFolder( ) + "/metaData.data" ) );

		return loadUtil.loadResultClass( );
	}
	
	/**
	 * @return
	 */
	private String getSaveFolder()
	{
		return getDataSetCacheManager( ).getSaveFolder( );
	}

	/**
	 * 
	 * @return
	 */
	private DataSetCacheManager getDataSetCacheManager( )
	{
		return this.session.getDataSetCacheManager( );
	}
	
	/**
	 * @return
	 */
	private int getCacheCount()
	{
		return getDataSetCacheManager( ).getCacheRowCount( );
	}
	
	/** 
	 * @return
	 */
	private String getLoadFolder()
	{
		return getDataSetCacheManager( ).getLoadFolder( );
	}
	
	/**
	 * Util class to save the original data retrieved from ODA driver into cache
	 * file.
	 */
	private class SaveUtil
	{
		private File file;
		private File metaFile;
		private FileOutputStream fos;
		private BufferedOutputStream bos;
		
		private IResultClass rsClass;
		private ResultObjectUtil roUtil;
		
		private int rowCount;
		
		/**
		 * @param file
		 * @param rsClass
		 */
		private SaveUtil( File file, File metaFile, IResultClass rsClass )
		{
			assert file != null;
			assert metaFile != null;
			assert rsClass != null;

			this.file = file;
			this.file.deleteOnExit( );
			this.metaFile = metaFile;
			this.metaFile.deleteOnExit( );
			this.rsClass = rsClass;
			this.rowCount = 0;
		}
		
		/**
		 * @param resultObject
		 * @throws DataException
		 */
		private void saveObject( IResultObject resultObject ) throws DataException
		{
			assert resultObject!=null;
			
			if ( roUtil == null )
			{				
				roUtil = ResultObjectUtil.newInstance( rsClass );
				try
				{
					fos = new FileOutputStream(file);
					bos = new BufferedOutputStream( fos );
				}
				catch ( FileNotFoundException e )
				{
					throw new DataException( ResourceConstants.DATASETCACHE_SAVE_ERROR,
							e );
				}
			}
			
			try
			{
				rowCount ++;
				roUtil.writeData( bos, resultObject );
			}
			catch ( IOException e )
			{
				throw new DataException( ResourceConstants.DATASETCACHE_SAVE_ERROR,
						e );
			}
		}
		
		/**
		 * @throws DataException
		 */
		private void close( ) throws DataException
		{			
			try
			{
				if ( bos != null )
				{
					bos.close( );
					fos.close( );
				}
				
				FileOutputStream fos1 = new FileOutputStream( metaFile );
				BufferedOutputStream bos1 = new BufferedOutputStream( fos1 );

				// save the count of data
				IOUtil.writeInt( bos1, this.rowCount );
				// save the meta data of result
				Map metaMap = new HashMap( );
				populateDataSetRowMapping( metaMap );
				( (ResultClass) rsClass ).doSave( bos1, metaMap );

				bos1.close( );
				fos1.close( );
			}
			catch ( IOException e )
			{
				throw new DataException( ResourceConstants.DATASETCACHE_SAVE_ERROR,
						e );
			}			
		}

		/**
		 * 
		 * @param metaMap
		 * @throws DataException
		 */
		private void populateDataSetRowMapping( Map metaMap )
				throws DataException
		{
			for ( int i = 0; i < rsClass.getFieldCount( ); i++ )
				metaMap.put( rsClass.getFieldName( i + 1 ),
						new ScriptExpression( ExpressionUtil.createJSDataSetRowExpression( rsClass.getFieldName( i + 1 ) ) ) );
		}
	}
	
	/**
	 * Util class to retrieve data from stored cache file.
	 */
	private class LoadUtil
	{
		private File file;
		private File metaFile;
		
		private FileInputStream fis;
		private BufferedInputStream bis;
		
		private ResultObjectUtil roUtil;
		private IResultClass rsClass;
		private int rowCount;
		private int currIndex;
		
		/**
		 * @param file
		 * @param metaFile
		 */
		private LoadUtil( File file, File metaFile )
		{
			assert file != null;
			assert metaFile != null;
			 
			this.file = file;
			this.metaFile = metaFile;
			this.rowCount = 0;
			this.currIndex = -1;
		}
		
		/**
		 * @param resultObject
		 * @throws DataException
		 */
		private IResultObject loadObject( ) throws DataException
		{			
			if ( roUtil == null )
				init( );
			
			try
			{
				if ( currIndex == rowCount - 1 )
					return null;
				
				currIndex++;
				return roUtil.readData( bis, 1 )[0];
			}
			catch ( IOException e )
			{
				throw new DataException( ResourceConstants.DATASETCACHE_LOAD_ERROR,
						e );
			}
		}
		
		/**
		 * @return
		 * @throws DataException
		 */
		private IResultClass loadResultClass( ) throws DataException
		{			
			if ( roUtil == null )
				init( );
			
			return this.rsClass;
		}
		
		/**
		 * @throws DataException
		 */
		private void init( ) throws DataException
		{
			try
			{
				FileInputStream fis1 = new FileInputStream( metaFile );
				BufferedInputStream bis1 = new BufferedInputStream( fis1 );

				rowCount = IOUtil.readInt( bis1 );
				rsClass = new ResultClass( bis1 );

				bis1.close( );
				fis1.close( );

				if ( rowCount > 0 )
				{
					roUtil = ResultObjectUtil.newInstance( rsClass );
					fis = new FileInputStream( file );
					bis = new BufferedInputStream( fis );
				}
			}
			catch ( FileNotFoundException e )
			{
				throw new DataException( ResourceConstants.DATASETCACHE_LOAD_ERROR,
						e );
			}
			catch ( IOException e )
			{
				throw new DataException( ResourceConstants.DATASETCACHE_LOAD_ERROR,
						e );
			}
		}
		
		/**
		 * @throws DataException
		 */
		private void close( ) throws DataException
		{
			if ( bis != null )
			{
				try
				{
					bis.close( );
					fis.close( );
				}
				catch ( IOException e )
				{
					// ignore;
				}
			}
		}
	}
	
}
