/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.data.engine.impl.document;

import java.io.OutputStream;

import org.eclipse.birt.core.archive.RAInputStream;
import org.eclipse.birt.data.engine.api.DataEngineContext;
import org.eclipse.birt.data.engine.core.DataException;

/**
 * Manage the input stream and output stream when reading/writing data into
 * report document. There is a big difference for queryResultID in
 * reading/writing. In reading context, the rootQueryResultID,
 * parentQueryResultID and queryResultID will be combined to get the correct
 * input stream. But in writing context, only the queryResultID is needed.
 */
public class StreamManager
{

	/**
	 * A possible directory of stream:
	 * 
	 * [QuRs0] ----------------------------------<<QUERY_ROOT_STREAM>>, <<BASE_SCOPE>>
	 * 		DataSetData
	 * 		ExprMetaInfo
	 * 		ExprValue
	 * 		QueryDefn
	 * 		GroupInfo
	 * 		QueryIDInfo
	 * 		ResultClass
	 * 		RowLengthInfo
	 * -----[IAMTEST] ---------------------------<<SUB_QUERY_ROOT_STREAM>>
	 * 			SubQueryInfo
	 * ---------[0]   ---------------------------<<SUB_QUERY_STREAM>>
	 * 				ExprMetaInfo
	 * 				ExprValue
	 * 				GroupInfo
	 * 				RowLengthInfo
	 * ---------[1]
	 * 				ExprMetaInfo
	 * 				ExprValue
	 * 				GroupInfo
	 * 				RowLengthInfo
	 * -----[QuRs1] -----------------------------<<QUERY_ROOT_STREAM>>, <<SELF_SCOPE>>
	 * 			QueryDefn
	 * 			GroupInfo
	 * 			RowIndexInfo
	 * ---------[IAMTEST] -----------------------<<SUB_QUERY_ROOT_STREAM>>
	 * 				SubQueryInfo
	 * -------------[0]   -----------------------<<SUB_QUERY_STREAM>>
	 * 					GroupInfo                
	 * 					RowIndexInfo             << maybe refer to 1 of parent >>
	 * -------------[1]
	 * 					GroupInfo
	 * 					RowIndexInfo	         << maybe refer to 0 of parent >>
	 * 
	 * */
	
	/**
	 * Actually RDLoad and RDSave needs to have a statue to indicate which type
	 * of query is running, and possible values are:
	 * 
	 * Normal query.
	 * Normal sub query.
	 * 
	 * Query is generated from normal query. 
	 * Sub query is generated from normal sub query.
	 */
	
	//
	private DataEngineContext context;
	
	private String rootQueryResultID;
	private String parentQueryResultID;
	private String selfQueryResultID;

	private String subQueryID;
	private String subQueryName;
	
	public final static int BASE_SCOPE = 0;
	public final static int PARENT_SCOPE = 1;
	public final static int SELF_SCOPE = 2;
	
	public final static int ROOT_STREAM = 0;
	public final static int SUB_QUERY_ROOT_STREAM = 1;
	public final static int SUB_QUERY_STREAM = 2;
	
	/**
	 * @param context
	 */
	public StreamManager( DataEngineContext context, QueryResultInfo queryResultInfo )
	{
		this.context = context;
		
		this.rootQueryResultID = queryResultInfo.getRootQueryResultID( );
		this.parentQueryResultID = queryResultInfo.getParentQueryResultID( );
		this.selfQueryResultID = queryResultInfo.getSelfQueryResultID( );
		
		this.subQueryName = queryResultInfo.getSubQueryName( );
		this.subQueryID = subQueryName == null ? null
				: QueryResultIDUtil.buildSubQueryID( subQueryName,
						queryResultInfo.getIndex( ) );
	}
	
	/**
	 * @param streamType
	 * @param streamPos
	 * @param streamScope
	 * @return
	 * @throws DataException
	 */
	OutputStream getOutStream( int streamType, int streamPos, int streamScope )
			throws DataException
	{
		StreamID streamID = getStreamID( streamPos, streamScope );
		return context.getOutputStream( streamID.startStream,
				streamID.subQueryStream,
				streamType );
	}
	
	/**
	 * @param streamType
	 * @return
	 * @throws DataException
	 */
	public RAInputStream getInStream( int streamType, int streamPos, int streamScope ) throws DataException
	{
		StreamID streamID = getStreamID( streamPos, streamScope );
		return (RAInputStream) context.getInputStream( streamID.startStream,
				streamID.subQueryStream,
				streamType );
	}
	
	/**
	 * This logic is special for sub query, since it needs to dynamically change
	 * to locate its parent sub query folder. The reason is the sub query indx
	 * in the new sub query might not be the same as its original sub query
	 * index. Don't use it in other cases.
	 * 
	 * @param streamType
	 * @param streamPos
	 * @param streamScope
	 * @return the input stream for sub query index of original sub query
	 * @throws DataException
	 */
	public RAInputStream getInStream2( int streamType, int streamPos,
			int streamScope, int parentIndex ) throws DataException
	{
		String tempID = subQueryID;
		subQueryID = QueryResultIDUtil.buildSubQueryID( subQueryName, parentIndex );
		StreamID streamID = getStreamID( streamPos, streamScope );
		subQueryID = tempID;
		
		return (RAInputStream) context.getInputStream( streamID.startStream,
				streamID.subQueryStream,
				streamType );
	}
	
	/**
	 * @param streamType
	 * @return
	 * @throws DataException
	 */
	boolean hasInStream( int streamType, int streamPos, int streamScope )
			throws DataException
	{
		StreamID streamID = getStreamID( streamPos, streamScope );
		return context.hasInStream( streamID.startStream,
				streamID.subQueryStream,
				streamType );
	}
	
	/**
	 * @param streamType
	 * @return
	 */
	public boolean hasOutStream( int streamType, int streamPos, int streamScope )
	{
		StreamID streamID = getStreamID( streamPos, streamScope );
		return context.hasOutStream( streamID.startStream,
				streamID.subQueryStream,
				streamType );
	}
	
	/**
	 * Drop specified streamType
	 * 
	 * @param streamType
	 * @throws DataException
	 */
	public void dropStream1( int streamType ) throws DataException
	{
		StreamID streamID = getStreamID( StreamManager.ROOT_STREAM,
				StreamManager.BASE_SCOPE );
		context.dropStream( streamID.startStream, null, streamType );
	}
	
	/**
	 * Drop specified sub stream
	 * 
	 * @param subStream
	 * @throws DataException
	 */
	void dropStream2( String subStream ) throws DataException
	{		
		StreamID streamID = getStreamID( StreamManager.ROOT_STREAM,
				StreamManager.BASE_SCOPE );
		String realStream = streamID.startStream + "/" + subStream;
		context.dropStream( realStream );
	}
	
	/**
	 * @param streamPos
	 * @param streamScope
	 * @return
	 */
	private StreamID getStreamID( int streamPos, int streamScope )
	{
		String startStream = null;
		if ( streamScope == BASE_SCOPE )
			if ( rootQueryResultID != null )
				startStream = rootQueryResultID;
			else
				startStream = selfQueryResultID;
		else if ( streamScope == PARENT_SCOPE )
			startStream = QueryResultIDUtil.getRealStreamID( rootQueryResultID,
					parentQueryResultID );
		else if ( streamScope == SELF_SCOPE )
			startStream = QueryResultIDUtil.getRealStreamID( rootQueryResultID,
					selfQueryResultID );

		String subQueryStream = null;
		if ( isSubquery( ) == false )
		{
			if ( streamPos == ROOT_STREAM )
				subQueryStream = null;
			else if ( streamPos == SUB_QUERY_ROOT_STREAM )
				subQueryStream = subQueryName;
			else if ( streamPos == SUB_QUERY_STREAM )
				subQueryStream = subQueryID;
		}
		else
		{
			if ( streamPos == ROOT_STREAM )
				subQueryStream = subQueryID;
			else if ( streamPos == SUB_QUERY_ROOT_STREAM )
				subQueryStream = subQueryName;
			else if ( streamPos == SUB_QUERY_STREAM )
				subQueryStream = subQueryID;
		}
		
		return new StreamID( startStream, subQueryStream );
	}	
	
	/**
	 * simple stream id wrapper 
	 */
	private static class StreamID
	{
		//
		private String startStream;
		private String subQueryStream;

		private StreamID( String startStream, String subQueryStream )
		{
			this.startStream = startStream;
			this.subQueryStream = subQueryStream;
		}
	}
	
	/**
	 * @return
	 */
	boolean isSubquery( )
	{
		return this.subQueryName != null;
	}
	
	/**
	 * Original query -> report document 1 
	 * 							
	 * new query ---------> report document 1.2
	 * 							
	 * new query ---------------------> report document 1.2.1
	 * 							
	 * new query ---------> report document 1.3
	 * 
	 * Used in read. To determine whether the associated stream manager is to
	 * read the query which is running on a report document.
	 * 
	 * @return
	 * @throws DataException
	 */
	boolean isBasedOnSecondRD( ) throws DataException
	{
		return parentQueryResultID != null;
	}
	
	/**
	 * Used in read and write. To determine whether the associated stream
	 * manager is to read/write the query which is running on a report document.
	 * If isBasedOnSecondRD is true, isSecondRD must be true.
	 * 
	 * @return
	 * @throws DataException
	 */
	boolean isSecondRD( ) throws DataException
	{
		return rootQueryResultID != null;
	}
	
	/**
	 * @return
	 */
	String getQueryResultUID( )
	{
		return QueryResultIDUtil.buildID( rootQueryResultID, selfQueryResultID );
	}

}
