/*
 *************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *  
 *************************************************************************
 */

package org.eclipse.birt.data.oda.util.manifest;

import java.util.HashMap;
import org.eclipse.birt.core.framework.IConfigurationElement;
import org.eclipse.birt.data.oda.OdaException;
import org.eclipse.birt.data.oda.util.logging.Level;

/**
 * Configuration of the driver's trace logging settings
 * for a data source extension.
 * @deprecated	As of BIRT 2.0, replaced by
 * 		{@link org.eclipse.datatools.connectivity.oda.util.manifest.TraceLogging 
 * 				org.eclipse.datatools.connectivity.oda.util.manifest.TraceLogging} .
 * 		<br>The BIRT ODA run-time framework has been migrated to the 
 * 		Eclipse Data Tools Platform (DTP) project.
 */
public class TraceLogging
{
	private int m_logLevel;
	private String m_logFileNamePrefix;
	private String m_logDirectory;
	private String m_logFormatterClass;
    private static HashMap sm_logLevelLiterals = null;
	
	TraceLogging( IConfigurationElement traceLogging, String dataSourceElementId ) throws OdaException
	{
		String logLevelString = traceLogging.getAttribute( "logLevel" );
		m_logLevel = toLogLevelNumber( logLevelString );
		
		m_logFileNamePrefix = traceLogging.getAttribute( "logFileNamePrefix" );		
		m_logDirectory = traceLogging.getAttribute( "logDirectory" );
		m_logFormatterClass = traceLogging.getAttribute( "logFormatterClass" );
	}
	
	/**
	 * Returns the value for the driver's log level,
	 * as defined in <code>org.eclipse.birt.data.oda.util.logging.Level</code>.
	 * @return	the log level.
	 */
	public int getLogLevel()
	{
		return m_logLevel;
	}
	
	/**
	 * Returns the string prefix for driver's log file names.
	 * @return	the log file name prefix.
	 */
	public String getLogFileNamePrefix()
	{
		return m_logFileNamePrefix;
	}
	
	/**
	 * Returns the optional directory for log files.
	 * @return	the log directory, or null if a log directory was not specified.
	 */
	public String getLogDirectory()
	{
		return m_logDirectory;
	}
	
	/**
	 * Returns the optional driver's concrete implementation of the 
	 * <i>org.eclipse.birt.data.oda.util.logging.LogFormatter</i> class.
	 * @return	the fully qualified class name for the log formatter class, or 
	 * 			null if the class was not specified.
	 */
	public String getLogFormatterClass()
	{
		return m_logFormatterClass;
	}
    
    /**
     * Converts a string that represents a log level name or 
     * numeric value to a number.  
     * Returns the default WARNING log level if given string value is null or
     * not valid.
     * @param logLevelName		A string that represents a log level name or numeric value.
     * @return                  A log level number.
     */
    static public int toLogLevelNumber( String logLevelName ) 
    {
        int defaultLogLevel = Level.WARNING;
        if( logLevelName == null || logLevelName.length() == 0 )
            return defaultLogLevel;
        
        // first check if given string value is a pre-defined log level name or number
        Integer mappedLevel = (Integer) getLogLevelLiterals().get( logLevelName );
        if( mappedLevel != null )   // found a match
            return mappedLevel.intValue();
        
        // could be numeric string value, try to convert to number
        try
        {
            int logLevel = Short.parseShort( logLevelName );
    		if( logLevel > Level.SEVERE )
    		    logLevel = Level.OFF;
    		return logLevel;
        }
        catch( NumberFormatException ex )
        {
            return defaultLogLevel;
        }
    }
    
    static private HashMap getLogLevelLiterals()
    {
        if( sm_logLevelLiterals != null )
            return sm_logLevelLiterals;
        
        sm_logLevelLiterals = new HashMap( 18, 1 );
        sm_logLevelLiterals.put( "SEVERE", new Integer( Level.SEVERE ) );
        sm_logLevelLiterals.put( "1000", new Integer( Level.SEVERE ) );
        sm_logLevelLiterals.put( "WARNING", new Integer( Level.WARNING ) );
        sm_logLevelLiterals.put( "900", new Integer( Level.WARNING ) );
        sm_logLevelLiterals.put( "INFO", new Integer( Level.INFO ) );
        sm_logLevelLiterals.put( "800", new Integer( Level.INFO ) );
        sm_logLevelLiterals.put( "CONFIG", new Integer( Level.CONFIG ) );
        sm_logLevelLiterals.put( "700", new Integer( Level.CONFIG ) );
        sm_logLevelLiterals.put( "FINE", new Integer( Level.FINE ) );
        sm_logLevelLiterals.put( "500", new Integer( Level.FINE ) );
        sm_logLevelLiterals.put( "FINER", new Integer( Level.FINER ) );
        sm_logLevelLiterals.put( "400", new Integer( Level.FINER ) );
        sm_logLevelLiterals.put( "FINEST", new Integer( Level.FINEST ) );
        sm_logLevelLiterals.put( "300", new Integer( Level.FINEST ) );
        sm_logLevelLiterals.put( "ALL", new Integer( Level.ALL ) );
        sm_logLevelLiterals.put( "0", new Integer( Level.ALL ) );
        sm_logLevelLiterals.put( "OFF", new Integer( Level.OFF ) );
        return sm_logLevelLiterals;
    }
}
