/*
 *******************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation. All rights reserved. This program and
 * the accompanying materials are made available under the terms of the Eclipse
 * Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: Actuate Corporation - initial API and implementation
 ******************************************************************************
 */

package org.eclipse.birt.report.data.oda.jdbc;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

/**
 * The SQL text formatter to escape comments and '\n' charactor. There are two
 * different types of comments: the --comment and the \/*...*\/ comment.
 * Comments can be insert into one seperate line, nested at the end of a SQL
 * command line,or with in the sql statement. This sql comments will be never
 * evaluated by the jdbc driver
 */
public class SQLFormatter
{
	// End of line
	private static final int CHAR_EOL = 0;
	// Comments
	private static final int CHAR_COMMENTS = 1;
	// SQL text
	private static final int CHAR_SQL_STR = 2;

	/**
	 * format the query text to escape the comments line and '\n' charactor
	 * 
	 * @param text
	 * @return
	 * @throws IOException
	 */
	public static String formatQueryText( String text )
	{
		if ( text == null )
			return null;
		List tokens;
		try
		{
			tokens = parse( text );
		}
		catch ( IOException e )
		{
			return text;
		}

		StringBuffer textBuffer = new StringBuffer( );
		for ( int i = 0; i < tokens.size( ); i++ )
		{
			if ( ( (SQLToken) tokens.get( i ) ).getType( ) == CHAR_EOL )
				textBuffer.append( ' ' );
			else if ( ( (SQLToken) tokens.get( i ) ).getType( ) != CHAR_COMMENTS )
				textBuffer.append( ( (SQLToken) tokens.get( i ) ).getValue( ) );
		}
		return textBuffer.toString( );
	}

	/**
	 * parse the query text into SQLToken.
	 * @param text
	 * @return
	 * @throws IOException 
	 */
	private static List parse( String text ) throws IOException
	{
		List tokenList = new ArrayList( );

		StringReader reader = new StringReader( text );

		try
		{
			int i;
			int startPosition = 0;
			int endPosition = -1;
			StringBuffer sqlText = new StringBuffer( );
			while ( ( i = reader.read( ) ) != -1 )
			{
				if ( i == '\n' )
				{
					startPosition = saveSQLText( tokenList,
							startPosition,
							sqlText );
					tokenList.add( new SQLToken( CHAR_EOL, '\n', startPosition ) );
					startPosition++;
				}
				else if ( i == '\"' || i == '\'' )
				{
					sqlText.append( (char) i );
					readNextQuote( sqlText, reader, i );
				}
				else if ( i == '-' )
				{
					int next = reader.read( );
					if ( next == '-' )
					{
						startPosition = saveSQLText( tokenList,
								startPosition,
								sqlText );
						StringBuffer value = readHyphensComments( reader );
						endPosition = startPosition + value.length( ) + 2;
						tokenList.add( new SQLToken( CHAR_COMMENTS,
								value.toString( ),
								startPosition,
								endPosition ) );
						startPosition = endPosition;
					}
					else
					{
						sqlText.append( (char) i );
						sqlText.append( (char) next );
					}
				}
				else if ( i == '/' )
				{
					int next = reader.read( );
					if ( next == '*' )
					{
						StringBuffer cmtBuf = new StringBuffer( );
						if ( isMultiLineComments( cmtBuf, reader ) )
						{
							startPosition = saveSQLText( tokenList,
									startPosition,
									sqlText );
							endPosition = startPosition
									+ cmtBuf.toString( ).length( ) + 2;
							tokenList.add( new SQLToken( CHAR_COMMENTS,
									cmtBuf.toString( ),
									startPosition,
									endPosition ) );
							startPosition = endPosition;
						}
						else
						{
							sqlText.append( "/" );
							sqlText.append( "*" );
							sqlText.append( cmtBuf.toString( ) );
						}
					}
					else
					{
						sqlText.append( "/" );
						sqlText.append( (char) next );
					}
				}
				else
				{
					sqlText.append( (char)i );
				}
			}
			saveSQLText( tokenList, startPosition, sqlText );
		}
		catch ( IOException e )
		{
			e.printStackTrace( );
		}
		
		return tokenList;
	}

	/**
	 * save the sql text into token list
	 * @param tokenList
	 * @param startPosition
	 * @param sqlText
	 * @return
	 */
	private static int saveSQLText( List tokenList, int startPosition,
			StringBuffer sqlText )
	{
		if ( sqlText.toString( ).length( ) > 0 )
		{
			tokenList.add( new SQLToken( CHAR_SQL_STR,
					sqlText.toString( ),
					startPosition,
					startPosition + sqlText.toString( ).length( ) ) );
			startPosition = startPosition + sqlText.toString( ).length( );
			sqlText.delete( 0, sqlText.toString( ).length( ) );
		}
		return startPosition;
	}

	/**
	 * get --comments buffer
	 * @param reader
	 * @return
	 * @throws IOException
	 */
	private static StringBuffer readHyphensComments( StringReader reader )
			throws IOException
	{
		int i = -1;
		StringBuffer str = new StringBuffer( );
		while ( ( i = reader.read( ) ) != '\n' && i != -1 )
		{
			str.append( (char) i );
		}
		return str;
	}

	/**
	 * get \/**\/ comments buffer
	 * 
	 * @param cmtBuf
	 * @param reader
	 * @return
	 * @throws IOException
	 */
	private static boolean isMultiLineComments( StringBuffer cmtBuf,
			StringReader reader ) throws IOException
	{
		int i = -1;
		boolean isComments = false;
		while ( ( i = reader.read( ) ) != -1 )
		{
			if ( i == '*' )
			{
				int next = reader.read( );
				if ( next == '/' )
				{
					isComments = true;
					break;
				}
				else
				{
					cmtBuf.append( (char) i );
					cmtBuf.append( (char) next );
				}
			}
			else
				cmtBuf.append( (char)i );
		}
		return isComments;
	}

	/**
	 * get the quoted string
	 * 
	 * @param strBuf
	 * @param reader
	 * @return
	 * @throws IOException
	 */
	private static StringBuffer readNextQuote( StringBuffer strBuf,
			StringReader reader, int quote ) throws IOException
	{
		int i = -1;
		while ( ( i = reader.read( ) ) != -1 )
		{
			if ( i != quote )
				strBuf.append( (char) i );
			else
			{
				strBuf.append( (char) quote );
				break;
			}
		}
		return strBuf;
	}
}

/**
 * Token symbol for sql text
 * 
 */
class SQLToken
{

	private int type;
	private int start;
	private int end;
	private String value;

	/**
	 * 
	 * @param type
	 * @param value
	 * @param start
	 */
	public SQLToken( int type, char value, int start )
	{
		this.type = type;
		this.value = ( new Character( value ) ).toString( );
		this.start = start;
		end = start + 1;
	}

	/**
	 * 
	 * @param type
	 * @param value
	 * @param start
	 * @param end
	 */
	public SQLToken( int type, String value, int start, int end )
	{
		this.type = type;
		this.value = value;
		this.start = start;
		this.end = end;
	}

	/**
	 * end position
	 * 
	 * @return
	 */
	public int getEnd( )
	{
		return end;
	}

	/**
	 * start position
	 * 
	 * @return
	 */
	public int getStart( )
	{
		return start;
	}

	/**
	 * token type
	 * 
	 * @return
	 */
	public int getType( )
	{
		return type;
	}

	/**
	 * the value
	 * 
	 * @return
	 */
	public String getValue( )
	{
		return value;
	}

	/**
	 * toString()
	 */
	public String toString( )
	{
		return value + "[" + type + "](" + start + "--" + end + ")";
	}
}
