/*******************************************************************************
 * Copyright (c) 2004, 2005 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.designer.data.ui.dataset;

import org.eclipse.birt.core.exception.BirtException;
import org.eclipse.birt.data.engine.api.DataEngine;
import org.eclipse.birt.data.engine.api.DataEngineContext;
import org.eclipse.birt.data.engine.api.IBaseDataSetDesign;
import org.eclipse.birt.data.engine.api.IBaseDataSourceDesign;
import org.eclipse.birt.report.designer.data.ui.property.AbstractDescriptionPropertyPage;
import org.eclipse.birt.report.designer.nls.Messages;
import org.eclipse.birt.report.model.api.DataSetHandle;
import org.eclipse.birt.report.model.api.PropertyHandle;
import org.eclipse.birt.report.model.api.activity.SemanticException;
import org.eclipse.jface.dialogs.IMessageProvider;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

/**
 * Cache preference set page. If the checkbox is selected, there are two optinal
 * ways to define the size of cache. User can assign the size of the cache or
 * select the entire data set as cache size. The selection will work on the data
 * set selection.
 */
public class CachePreferencePage extends AbstractDescriptionPropertyPage
{
	private boolean selected = false;
	private boolean radioSelected = false;
	
	private int rowCount = 0;
	private int preCount = rowCount;
	private int initialCount = rowCount;
	//cache the entire row set
	private final int ENTIRE_ROW_CACHE = -1;
	//not use the cache
	private final int UNUSED_ROW_CACHE = 0;
	private transient PropertyHandle rowCountHandle = null;
	private transient IBaseDataSetDesign dataSetDesign;
	private transient IBaseDataSourceDesign dataSourceDesign;

	private static String DEFAULT_MESSAGE = Messages.getString( "dataset.editor.cachePreference" );
	
	/**
	 * create Contents
	 */
	public Control createContents( Composite parent )
	{
		Composite composite = new Composite( parent, SWT.NONE );

		GridLayout layout = new GridLayout( );
		layout.numColumns = 5;
		composite.setLayout( layout );
		composite.setLayoutData( new GridData( GridData.HORIZONTAL_ALIGN_FILL
				| GridData.VERTICAL_ALIGN_FILL ) );

		Button previewNumCheckBox = new Button( composite, SWT.CHECK );
		GridData data = new GridData( );
		data.horizontalSpan = 5;
		previewNumCheckBox.setLayoutData( data );
		previewNumCheckBox.setText( Messages.getString( "CachePreferencePage.checkBox.text" ) );

		data = new GridData( );
		data.horizontalSpan = 2;
		Label lab = new Label( composite, SWT.NONE );
		lab.setLayoutData( data );
		final Button numRadioBox = new Button( composite, SWT.RADIO );
		numRadioBox.setText( Messages.getString( "CachePreferencePage.radioBox.numText" ) );
		numRadioBox.setEnabled( false );

		data = new GridData( );
		data.widthHint = 50;
		final Text numText = new Text( composite, SWT.BORDER );
		numText.setEnabled( false );
		numText.setLayoutData( data );
		numText.setData( rowCountHandle );

		Label label = new Label( composite, SWT.NONE );
		label.setText( Messages.getString( "CachePreferencePage.label.text" ) );

		data = new GridData( );
		data.horizontalSpan = 2;
		Label lab1 = new Label( composite, SWT.NONE );
		lab1.setLayoutData( data );
		data = new GridData( );
		data.horizontalSpan = 3;
		final Button entireNumRadioBox = new Button( composite, SWT.RADIO );
		entireNumRadioBox.setLayoutData( data );
		entireNumRadioBox.setEnabled( false );
		entireNumRadioBox.setText( Messages.getString( "CachePreferencePage.radioBox.entireNumText" ) );

		data = new GridData( );
		data.horizontalSpan = 3;
		final Button clearButton = new Button( composite, SWT.NONE );
		clearButton.setText( Messages.getString( "CachePreferencePage.button.clear" ) );
		clearButton.setLayoutData( data );

		previewNumCheckBox.addSelectionListener( new SelectionAdapter( ) {

			public void widgetSelected( SelectionEvent e )
			{
				selected = !selected;
				numRadioBox.setEnabled( selected );
				entireNumRadioBox.setEnabled( selected );
				if ( !selected )
				{
					getContainer( ).setMessage( DEFAULT_MESSAGE,
							IMessageProvider.NONE );
					numText.setEnabled( selected );
					rowCount = UNUSED_ROW_CACHE;
				}
				else
				{
					if ( radioSelected )
					{
						numText.setEnabled( selected );
						setRowCount( numText.getText( ) );
					}
					else
					{
						entireNumRadioBox.setSelection( selected );
						numRadioBox.setSelection( radioSelected );
						rowCount = ENTIRE_ROW_CACHE;
					}
				}
			}

			public void widgetDefaultSelected( SelectionEvent e )
			{
			}
		} );

		numRadioBox.addSelectionListener( new SelectionAdapter( ) {

			public void widgetSelected( SelectionEvent e )
			{
				getContainer( ).setMessage( DEFAULT_MESSAGE,
						IMessageProvider.NONE );
				radioSelected = numRadioBox.getSelection( );
				numText.setEnabled( radioSelected );
				setRowCount( numText.getText( ) );
			}

			public void widgetDefaultSelected( SelectionEvent e )
			{
			}
		} );

		entireNumRadioBox.addSelectionListener( new SelectionAdapter( ) {

			public void widgetSelected( SelectionEvent e )
			{
				getContainer( ).setMessage( DEFAULT_MESSAGE,
						IMessageProvider.NONE );
				if ( entireNumRadioBox.getSelection( ) )
					rowCount = ENTIRE_ROW_CACHE;
			}

			public void widgetDefaultSelected( SelectionEvent e )
			{
			}
		} );

		numText.addModifyListener( new ModifyListener( ) {

			public void modifyText( ModifyEvent e )
			{
				getContainer( ).setMessage( DEFAULT_MESSAGE,
						IMessageProvider.NONE );
				setRowCount( numText.getText( ) );
			}
		} );

		clearButton.addSelectionListener( new SelectionAdapter( ) {

			// TODO add the selection event
			public void widgetSelected( SelectionEvent e )
			{
				if ( clearCache( ) )
				{
					MessageDialog.openInformation( null,
							Messages.getString( "CachePreferencePage.messageBox.title" ),
							Messages.getString( "CachePreferencePage.messageBox.info" ) );
				    clearButton.setEnabled( false );
				}
			}

			public void widgetDefaultSelected( SelectionEvent e )
			{
			}
		} );

		initCachePreference( );
		if ( rowCount > 0 )
		{
			selected = true;
			previewNumCheckBox.setSelection( selected );
			numRadioBox.setEnabled( true );
			numRadioBox.setSelection( true );
			numText.setEnabled( true );
			numText.setText( Integer.toString( rowCount ) );
			entireNumRadioBox.setEnabled( true );
		}
		else if ( rowCount == 0 )
		{
			previewNumCheckBox.setSelection( false );
		}
		else if ( rowCount < 0 )
		{
			previewNumCheckBox.setSelection( true );
			entireNumRadioBox.setSelection( true );
			selected = true;
			previewNumCheckBox.setSelection( selected );
			numRadioBox.setEnabled( true );
			entireNumRadioBox.setEnabled( true );
		}
		return composite;
	}
	
	/**
	 * set row count
	 * @param count
	 */
	private void setRowCount( String count )
	{
		try
		{
			rowCount = Integer.parseInt( count );
		}
		catch ( NumberFormatException ex )
		{
			getContainer( ).setMessage( Messages.getString( "dataset.editor.cachePreference.errorNumberFormat" ),
					IMessageProvider.ERROR );
			rowCount = 0;
		}
		if ( rowCount <= 0 )
		{
			getContainer( ).setMessage( Messages.getString( "dataset.editor.cachePreference.errorNumberFormat" ),
					IMessageProvider.ERROR );
		}
	}

	/**
	 * get the rowCount
	 * 
	 */
	private void initCachePreference( )
	{
		rowCount = ( (DataSetHandle) getContainer( ).getModel( ) ).getCachedRowCount( );
		rowCountHandle = ( (DataSetHandle) getContainer( ).getModel( ) ).getPropertyHandle( DataSetHandle.CACHED_ROW_COUNT_PROP );
		preCount = rowCount;
		initialCount = rowCount;
	}

	/**
	 * clear the cache
	 *
	 */
	public boolean clearCache( )
	{
		try
		{
			DataEngine dataEngine = DataEngine.newDataEngine( DataEngineContext.newInstance( DataEngineContext.DIRECT_PRESENTATION,
					null,
					null,
					null ) );
			dataEngine.clearCache( this.dataSourceDesign, this.dataSetDesign );
		}
		catch ( BirtException e )
		{
			// TODO:
			e.printStackTrace( );
			return true;
		}
		return true;
	}
	
	/**
	 * clear the cache when the ok pressed. When there's no change in cache
	 * size, do nothing. Or, clear the exsisting cache
	 * 
	 */
	public boolean clearCacheOnExit( )
	{
		if ( initialCount != rowCount )
		{
			return clearCache( );
		}
		else
			return true;
	}
	
	/**
	 * activate the preference page
	 */
	public void pageActivated( )
	{
		getContainer( ).setMessage( DEFAULT_MESSAGE, IMessageProvider.NONE );
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.ui.dialogs.properties.AbstractPropertyPage#performCancel()
	 */
	public boolean performCancel( )
	{
		// selectorImage.dispose( );
		return super.performCancel( );
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.ui.dialogs.properties.AbstractPropertyPage#canLeave()
	 */
	public boolean canLeave( )
	{
		if ( canFinish( ) )
			return super.performOk( );
		else
			return false;
	}

	/**
	 * whether the page can be finished or leave.
	 */
	public boolean canFinish( )
	{
		if ( ( selected && rowCount == 0 )
				|| ( selected && radioSelected && rowCount <= 0 ) )
		{
			getContainer( ).setMessage( Messages.getString( "dataset.editor.cachePreference.errorNumberFormat" ),
					IMessageProvider.ERROR );
			return false;
		}
		if ( preCount != rowCount && rowCountHandle != null )
		{
			try
			{
				this.rowCountHandle.clearValue( );
				this.rowCountHandle.setIntValue( rowCount );
				preCount = rowCount;
			}
			catch ( SemanticException e )
			{
				e.printStackTrace( );
			}
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.ui.dialogs.properties.AbstractPropertyPage#performOk()
	 */
	public boolean performOk( )
	{
		if ( canFinish( ) )
		{
			return super.performOk( );
		}
		else
			return false;
	}
	
	/**
	 * set dataSetDesign and dataSourceDesign
	 * 
	 * @param baseDataSetDesign
	 * @param baseDataSourceDesign
	 */
	public void setBaseDesign( IBaseDataSetDesign baseDataSetDesign,
			IBaseDataSourceDesign baseDataSourceDesign )
	{
		this.dataSetDesign = baseDataSetDesign;
		this.dataSourceDesign = baseDataSourceDesign;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.data.ui.property.AbstractPropertyPage#getToolTip()
	 */
	public String getToolTip( )
	{
		return Messages.getString( "CachePreferencePage.Filter.Tooltip" ); //$NON-NLS-1$
	}
} // @jve:decl-index=0:visual-constraint="351,164"
