/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.designer.internal.ui.views.attributes.widget;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import com.ibm.icu.util.ULocale;

import org.eclipse.birt.report.designer.core.model.SessionHandleAdapter;
import org.eclipse.birt.report.designer.ui.views.attributes.providers.ChoiceSetFactory;
import org.eclipse.birt.report.designer.ui.views.attributes.providers.PropertyProcessor;
import org.eclipse.birt.report.model.api.util.StringUtil;

/**
 * The descriptor to control resource files
 */

public class ResourceFileDescriptor extends ComboPropertyDescriptor
{

	private static final String RESOURCE_FILE_EXTENSION = ".properties"; //$NON-NLS-1$

	private static final String[] COUNTRIES = ULocale.getISOCountries( );

	private static final String[] LANGUAGES = ULocale.getISOLanguages( );

	private static boolean isValidIsoCountry( String cout )
	{
		if ( cout != null )
		{
			for ( int i = 0; i < COUNTRIES.length; i++ )
			{
				if ( COUNTRIES[i].equalsIgnoreCase( cout ) )
				{
					return true;
				}
			}
		}

		return false;
	}

	private static boolean isValidIsoLanguage( String lang )
	{
		if ( lang != null )
		{
			for ( int i = 0; i < LANGUAGES.length; i++ )
			{
				if ( LANGUAGES[i].equalsIgnoreCase( lang ) )
				{
					return true;
				}
			}
		}

		return false;
	}

	public ResourceFileDescriptor( PropertyProcessor propertyProcessor )
	{
		super( propertyProcessor );
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.birt.report.designer.internal.ui.views.attributes.widget.PropertyDescriptor#resetUIData()
	 */
	public void resetUIData( )
	{
		String fileName = SessionHandleAdapter.getInstance( )
				.getReportDesignHandle( )
				.getFileName( );
		File file = new File( fileName );
		String[] fileNames = file.getParentFile( ).list( new FilenameFilter( ) {

			public boolean accept( File arg0, String arg1 )
			{
				return arg1.endsWith( RESOURCE_FILE_EXTENSION );
			}
		} );
		List list = new ArrayList( );
		for ( int i = 0; i < fileNames.length; i++ )
		{
			String baseName = getBaseName( fileNames[i] );
			if ( baseName != null && !list.contains( baseName ) )
			{
				list.add( baseName );
			}
		}
		combo.setEnabled( true );
		oldValue = getPropertyProcessor( ).getStringValue( getModelList( ) );
		if ( list.isEmpty( ) )
		{
			if ( StringUtil.isBlank( oldValue ) )
			{
				combo.setEnabled( false );
			}
			else
			{
				combo.setText( oldValue );
			}
			return;
		}
		if(getPropertyProcessor( ).isReadOnly( getModelList( ) ))
		{
			combo.setEnabled(false);
		}
		
		Collections.sort( list );
		String[] items = new String[list.size( ) + 1];
		items[0] = ChoiceSetFactory.CHOICE_NONE;
		System.arraycopy( (String[]) list.toArray( new String[0] ),
				0,
				items,
				1,
				list.size( ) );
		combo.setItems( items );

		if ( !list.contains( oldValue ) )
		{
			combo.deselectAll( );
			if ( oldValue != null && oldValue.length( ) > 0 )
			{
				combo.setText( oldValue );
			}
			else
			{
				combo.select( 0 );
			}
			return;
		}
		combo.select( Arrays.asList( combo.getItems( ) ).indexOf( oldValue ) );
	}

	/**
	 * @param string
	 * @return
	 */
	private String getBaseName( String fileName )
	{
		if ( fileName.indexOf( " " ) != -1 ) //$NON-NLS-1$
		{//illegal filename
			return null;
		}
		fileName = fileName.substring( 0, fileName.length( )
				- RESOURCE_FILE_EXTENSION.length( ) );

		int index = fileName.lastIndexOf( "_" ); //$NON-NLS-1$

		if ( index > 0 )
		{
			String suffix = fileName.substring( index + 1 );

			int sep = suffix.indexOf( "-" ); //$NON-NLS-1$
			if ( sep != -1 )
			{
				String lang = suffix.substring( 0, sep );
				String cout = suffix.substring( sep + 1 );

				if ( isValidIsoCountry( cout ) && isValidIsoLanguage( lang ) )
				{
					return fileName.substring( 0, index );
				}
			}
			else if ( isValidIsoLanguage( suffix ) )
			{
				return fileName.substring( 0, index );
			}
			else if ( isValidIsoCountry( suffix ) )
			{
				int sIndex = fileName.lastIndexOf( "_", index - 1 ); //$NON-NLS-1$

				if ( sIndex == -1 )
				{
					return fileName.substring( 0, index );
				}
				else
				{
					String ss = fileName.substring( sIndex + 1, index );

					if ( isValidIsoLanguage( ss ) )
					{
						return fileName.substring( 0, sIndex );
					}
					else
					{
						return fileName.substring( 0, index );
					}
				}
			}
		}

		return fileName;
	}
}