/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.engine.data;

import java.math.BigDecimal;
import java.sql.Blob;
import java.util.Date;

import org.eclipse.birt.core.exception.BirtException;
import org.eclipse.birt.data.engine.api.IBaseExpression;
import org.eclipse.birt.data.engine.api.IResultMetaData;
import org.eclipse.birt.report.engine.api.DataSetID;

/**
 * Encapsulates a set of result rows. Provide functions to cycle through all
 * data rows
 * 
 */
public interface IResultSet
{
	/**
	 * get the parent result set,
	 * return NULL if it is the topmost query.
	 * @return
	 */
	IResultSet getParent();
	
	DataSetID getID();
	/**
	 * return the current position
	 * @return
	 */
	long getCurrentPosition();
	/**
	 * move the current position to the row with Id rows.
	 * @param rows row id start from 0.
	 * @return true if sucesseccful.
	 */
	boolean skipTo(long rows);
	/**
	 * Moves down one element from its current position of the ResultSet.
	 * 
	 * @return true if next element exists and has not reached the limit on the
	 *         maximum number of rows that can be accessed.
	 */
	boolean next();

	/**
	 * Returns the 1-based index of the outermost group in which the current row
	 * is the last row. For example, if a query contain N groups (group with
	 * index 1 being the outermost group, and group with index N being the
	 * innermost group), and this function returns a value M, it indicates that
	 * the current row is the last row in groups with indexes (M, M+1, ..., N ).
	 * -1 represents current row is a detail row. 0 represents the end of whole
	 * resultset
	 * 
	 * @return 1-based index of the outermost group in which the current row is
	 *         the first row; (N+1) if the current row is not at the start of
	 *         any group; 0 if the result set has no groups.
	 */
	int getStartingGroupLevel();

	/**
	 * Returns the 1-based index of the outermost group in which the current row
	 * is the first row. For example, if a query contain N groups (group with
	 * index 1 being the outermost group, and group with index N being the
	 * innermost group), and this function returns a value M, it indicates that
	 * the current row is the first row in groups with indexes (M, M+1, ..., N ).
	 * 
	 * @return 1-based index of the outermost group in which the current row is
	 *         the first row; (N+1) if the current row is not at the start of
	 *         any group; 0 if the result set has no groups.
	 */
	int getEndingGroupLevel();

    /**
     * Returns the metadata of this result set's detail row.
     * @return	The result metadata of a detail row.
     */
    public IResultMetaData getResultMetaData() throws BirtException;
    
    /**
	 * Returns the value of a bound column. Currently it is only a dummy
	 * implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Object getValue( String name ) throws BirtException;
    
    /**
	 * Returns the value of a bound column as the Boolean data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Boolean getBoolean( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the Integer data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Integer getInteger( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the Double data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Double getDouble( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the String data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public String getString( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the BigDecimal data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public BigDecimal getBigDecimal( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the Date data type. Currently
	 * it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Date getDate( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the Blob data type. Currently
	 * it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public Blob getBlob( String name ) throws BirtException;

    /**
	 * Returns the value of a bound column as the byte[] data type.
	 * Currently it is only a dummy implementation.
	 * 
	 * @param name of bound column
	 * @return value of bound column
	 * @throws BirtException
	 */
    public byte[] getBytes( String name ) throws BirtException;
    
	
	/**
	 * Closes this resultset
	 */
	void close();	
	
	/**
	 * Returns the value of a query result expression. A given data expression
	 * could be for one of the Selected Columns (if detail rows are used), or of
	 * an Aggregation specified in the prepared ReportQueryDefn spec. When
	 * requesting for the value of a Selected Column, its value in the current
	 * row of the iterator will be returned.
	 * <p>
	 * Throws an exception if a result expression value is requested out of
	 * sequence from the prepared ReportQueryDefn spec. E.g. A group aggregation
	 * is defined to be after_last_row. It would be out of sequence if requested
	 * before having iterated/skipped to the last row of the group. In future
	 * release, this could have intelligence to auto recover and performs
	 * dependent operations to properly evaluate any out-of-sequence result
	 * values.
	 * 
	 * @param dataExpr
	 *            An IExpression object provided in the ReportQueryDefn at the
	 *            time of prepare.
	 * @return The value of the given expression. It could be null.
	 */
	Object evaluate( String expr );
	
	Object evaluate( IBaseExpression expr );

	public String getGroupId( int groupLevel );

}
