/*******************************************************************************
 * Copyright (c) 2004 Actuate Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Actuate Corporation  - initial API and implementation
 *******************************************************************************/

package org.eclipse.birt.report.engine.presentation;

import java.util.Stack;

import org.eclipse.birt.report.engine.content.IBandContent;
import org.eclipse.birt.report.engine.content.ICellContent;
import org.eclipse.birt.report.engine.content.IContainerContent;
import org.eclipse.birt.report.engine.content.IContent;
import org.eclipse.birt.report.engine.content.IDataContent;
import org.eclipse.birt.report.engine.content.IForeignContent;
import org.eclipse.birt.report.engine.content.IImageContent;
import org.eclipse.birt.report.engine.content.ILabelContent;
import org.eclipse.birt.report.engine.content.IPageContent;
import org.eclipse.birt.report.engine.content.IReportContent;
import org.eclipse.birt.report.engine.content.IRowContent;
import org.eclipse.birt.report.engine.content.IStyle;
import org.eclipse.birt.report.engine.content.ITableBandContent;
import org.eclipse.birt.report.engine.content.ITableContent;
import org.eclipse.birt.report.engine.content.ITextContent;
import org.eclipse.birt.report.engine.emitter.ContentEmitterAdapter;
import org.eclipse.birt.report.engine.emitter.IContentEmitter;
import org.eclipse.birt.report.engine.emitter.IEmitterServices;
import org.eclipse.birt.report.engine.internal.document.DocumentExtension;
import org.eclipse.birt.report.engine.ir.MasterPageDesign;
import org.eclipse.birt.report.engine.ir.PageSetupDesign;
import org.eclipse.birt.report.engine.layout.ILayoutPageHandler;

public class DefaultPaginationEmitter extends ContentEmitterAdapter
{

	/**
	 * output emitter. The paginated content will pass out using this emitter.
	 */
	protected IContentEmitter emitter;

	/**
	 * page event handler.
	 */
	protected ILayoutPageHandler pageHandler;

	/**
	 * current page number.
	 */
	protected int currentPageNo = 0;
	/**
	 * first element in the page
	 */
	protected long startOffset = -1;
	/**
	 * last element in the page
	 */
	protected long endOffset = -1;

	/**
	 * the offset of the current output content
	 */
	protected long offset = -1;
	/**
	 * name of the current master page
	 */
	protected String masterPage;

	/** current page */
	protected Page page;

	/**
	 * current output content. For createPage(), it means we create a page and
	 * put the element into the new page. For cloesPage(), it means need output
	 * the content into the exist page and close the old page.
	 */
	protected IContent content;

	/** current page region */
	protected PageRegion pageRegion;

	/** stack used to save the last table drops */
	protected Stack tableDrops = new Stack( );

	/**
	 * builder which use this pagination emitter
	 */
	protected IPaginationBuilder builder;

	/**
	 * create a pagination emitter used by the builder.
	 * 
	 * @param builder
	 *            builder which use this emitter.
	 */
	DefaultPaginationEmitter( IPaginationBuilder builder )
	{
		this.builder = builder;
	}

	/**
	 * set the output emitter.
	 * 
	 * @param emitter
	 *            output emitter
	 */
	void setOutputEmitter( IContentEmitter emitter )
	{
		this.emitter = emitter;
	}

	/**
	 * set the page handler.
	 * 
	 * @param pageHandler
	 *            handler receive the event
	 */
	void setPageHandler( ILayoutPageHandler pageHandler )
	{
		this.pageHandler = pageHandler;
	}

	/**
	 * get the current page hint.
	 * 
	 * @return page hint.
	 */
	public IPageHint getPageHint( )
	{
		return new PageHint( currentPageNo, -1, startOffset, endOffset );
	}

	/**
	 * if the content is block element. page break can only occurs between the
	 * block elements.
	 * 
	 * @param content
	 * @return true block element, false not block element
	 */
	protected boolean isBlockLevelElement( IContent content )
	{
		IContent parent = (IContent) content.getParent( );
		if ( parent == null )
		{
			return true;
		}
		if ( isBlockLevelElement( parent ) )
		{
			IStyle style = content.getComputedStyle( );
			String display = style.getDisplay( );
			if ( IStyle.CSS_BLOCK_VALUE == display
					|| IStyle.CSS_TABLE_VALUE == display
					|| IStyle.CSS_TABLE_ROW_VALUE == display
					|| IStyle.CSS_TABLE_ROW_GROUP_VALUE == display )
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * check if the element is in table header. page-break can't occurs in the
	 * table header.
	 * 
	 * @param content
	 * @return true in the table header, false not in the table header
	 */
	protected boolean inRepeatHeader( IContent content )
	{
		if ( content instanceof IBandContent )
		{
			IBandContent band = (IBandContent) content;
			
			if ( band.getBandType( ) == IBandContent.BAND_HEADER
					|| band.getBandType( ) == IBandContent.GROUP_CONTENT )
			{
				return true;
			}
		}
		IContent parent = (IContent) content.getParent( );
		if ( parent != null )
		{
			return inRepeatHeader( parent );
		}
		return false;
	}

	/**
	 * if we can handle page-break in content. we can only handle page-break in:
	 * <li> block level element
	 * <li> if content is rows, it should not in table-header
	 * 
	 * @param content
	 *            content.
	 * @return
	 */
	protected boolean allowPageBreakBefore( IContent content )
	{
		if ( !tableDrops.isEmpty( ) )
		{
			DropStatus status = (DropStatus) tableDrops.peek( );
			if ( status.curRowId != status.dropRowId )
			{
				return false;
			}
		}
		return !inRepeatHeader( content ) ;
	}

	/**
	 * if we can handle page-break in content. we can only handle page-break in:
	 * <li> block level element
	 * <li> if content is rows, it should not in table-header
	 * 
	 * @param content
	 *            content.
	 * @return
	 */
	protected boolean allowPageBreakAfter( IContent content )
	{
		if ( !tableDrops.isEmpty( ) )
		{
			DropStatus status = (DropStatus) tableDrops.peek( );
			if ( status.curRowId + 1 != status.dropRowId )
			{
				return false;
			}
		}
		return !inRepeatHeader( content ) ;
	}

	/**
	 * get the output format of the emitter.
	 */
	public String getOutputFormat( )
	{
		if ( emitter != null )
		{
			return emitter.getOutputFormat( );
		}
		return null;
	}

	public void initialize( IEmitterServices service )
	{
		if ( emitter != null )
		{
			emitter.initialize( service );
		}
	}

	public void start( IReportContent report )
	{
		if ( emitter != null )
		{
			emitter.start( report );
		}
		PageSetupDesign pageSetup = report.getDesign( ).getPageSetup( );
		if(pageSetup.getMasterPageCount( )>0)
		{
			masterPage  = pageSetup.getMasterPage( 0 ).getName( );
		}
		// createDefaultPage( );
	}

	public void end( IReportContent report )
	{
		content = null;
		// termiate all pages if we have.
		closePage( );
		if ( currentPageNo == 0 )
		{
			// we always create a page for the empty report.
			createDefaultPage( );
			closePage( );
		}
		if ( emitter != null )
		{
			emitter.end( report );
		}
		if (pageHandler != null)
		{
			//notify that we have end all the pages
			LayoutContext context = new LayoutContext( currentPageNo,
					startOffset, endOffset, true );
			pageHandler.onPage( currentPageNo, context );
		}
		
	}

	public void startContainer( IContainerContent container )
	{
		handlePageBreakBefore( container );
		offset = getIndex( container );
		getCurrentPage( ).startContainer( container );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
	}

	public void endContainer( IContainerContent container )
	{
		getCurrentPage( ).endContainer( container );
		handlePageBreakAfter( container );
	}

	/**
	 * status control the drop area in the table.
	 * 
	 * we can't break the table between the rows which have drop areas.
	 * 
	 */
	private static class DropStatus
	{

		int curRowId;
		int dropRowId;
	}

	public void startTable( ITableContent table )
	{
		handlePageBreakBefore( table );
		DropStatus status = new DropStatus( );
		status.curRowId = 0;
		status.dropRowId = 0;
		tableDrops.push( status );
		offset = getIndex( table );
		getCurrentPage( ).startTable( table );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
	}

	public void endTable( ITableContent table )
	{
		tableDrops.pop( );
		getCurrentPage( ).endTable( table );
		handlePageBreakAfter( table );
	}

	public void startTableBand( ITableBandContent band )
	{
		handlePageBreakBefore( band );
		offset = getIndex( band );
		getCurrentPage( ).startTableBand( band );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
	}

	public void endTableBand( ITableBandContent band )
	{
		getCurrentPage( ).endTableBand( band );
		handlePageBreakAfter( band );
	}

	public void startRow( IRowContent row )
	{
		DropStatus status = (DropStatus) tableDrops.peek( );
		status.curRowId = row.getRowID( );
		handlePageBreakBefore( row );
		offset = getIndex( row );
		getCurrentPage( ).startRow( row );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
	}

	public void endRow( IRowContent row )
	{
		getCurrentPage( ).endRow( row );
		handlePageBreakAfter( row );
	}

	public void startCell( ICellContent cell )
	{
		DropStatus status = (DropStatus) tableDrops.peek( );
		int dropRowId = status.curRowId + cell.getRowSpan( );
		if ( dropRowId > status.dropRowId )
		{
			status.dropRowId = dropRowId;
		}
		offset = getIndex( cell );
		getCurrentPage( ).startCell( cell );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
	}

	public void endCell( ICellContent cell )
	{
		getCurrentPage( ).endCell( cell );
	}

	public void startText( ITextContent text )
	{
		handlePageBreakBefore( text );
		offset = getIndex( text );
		getCurrentPage( ).startText( text );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
		handlePageBreakAfter( text );
	}

	public void startLabel( ILabelContent label )
	{
		handlePageBreakBefore( label );
		offset = getIndex( label );
		getCurrentPage( ).startLabel( label );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
		handlePageBreakAfter( label );
	}

	public void startData( IDataContent data )
	{
		handlePageBreakBefore( data );
		offset = getIndex( data );
		getCurrentPage( ).startData( data );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
		handlePageBreakAfter( data );
	}

	public void startImage( IImageContent image )
	{
		handlePageBreakBefore( image );
		offset = getIndex( image );
		getCurrentPage( ).startImage( image );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
		handlePageBreakAfter( image );
	}

	public void startForeign( IForeignContent foreign )
	{
		handlePageBreakBefore( foreign );
		offset = getIndex( foreign );
		getCurrentPage( ).startForeign( foreign );
		if (startOffset == -1)
		{
			startOffset = offset;
		}
		handlePageBreakAfter( foreign );
	}

	private void handlePageBreakBefore( IContent content )
	{
		this.content = (IContent) content.getParent( );
		if ( allowPageBreakBefore( content ) )
		{
			if ( needPageBreakBefore( content ) )
			{
				createPage( );
			}
		}
	}

	protected boolean needPageBreakBefore( IContent content )
	{
		return hasMasterPageChanged( content );
	}

	protected boolean needPageBreakAfter( IContent content )
	{
		return false;
	}

	protected void handlePageBreakAfter( IContent content )
	{
		this.content = (IContent) content.getParent( );
		if ( allowPageBreakAfter( content ) )
		{
			if ( needPageBreakAfter( content ) )
			{
				closePage( );
			}
		}
	}

	/**
	 * create a new page, and set the current page to the new created.
	 * 
	 */
	protected void createPage( )
	{
		closePage( );
		startOffset = -1;
		IPageContent pageContent = createPageContent( );
		page = new Page( emitter, pageContent );
		page.open( );
		pageRegion = page.createRegion( );
		getCurrentPage( ).open( content );
	}

	/**
	 * return the current page, if the current page is NULL, create a new page
	 * and set it as curretn.
	 * 
	 * @return
	 */
	protected PageRegion getCurrentPage( )
	{
		if ( pageRegion == null )
		{
			createPage( );
		}
		assert pageRegion != null;
		return pageRegion;
	}

	static class LayoutContext
	{
		long currentPageNo;
		long startOffset;
		long endOffset;
		boolean finished;
		
		LayoutContext(long pageNumber, long start, long end, boolean f)
		{
			currentPageNo = pageNumber;
			startOffset = start;
			endOffset = end;
			finished = f;
		}
		
	};
	
	/**
	 * close the current page. After the close, the curent page is empty
	 */
	protected void closePage( )
	{
		endOffset = offset;
		if ( pageRegion != null )
		{
			pageRegion.close( content );
			pageRegion = null;
		}
		if ( page != null )
		{
			page.close( );
			page = null;
			if ( pageHandler != null )
			{
				LayoutContext context = new LayoutContext( currentPageNo,
						startOffset, endOffset, false );
				pageHandler.onPage( currentPageNo, context );
			}
		}
	}

	protected void createDefaultPage( )
	{
		createPage( );
	}

	protected IPageContent createPageContent( )
	{
		currentPageNo++;
		return builder.buildPageContent( currentPageNo, masterPage );
	}

	protected IContent clearBottomStyle( IContent content )
	{
		IStyle style = content.getStyle( );
		style.setProperty( IStyle.STYLE_MARGIN_TOP, IStyle.NONE_VALUE );
		style.setProperty( IStyle.STYLE_BORDER_TOP_STYLE, IStyle.NONE_VALUE );
		style.setProperty( IStyle.STYLE_PADDING_TOP, IStyle.NONE_VALUE );
		return content;
	}

	protected IContent cleartTopStyle( IContent content )
	{
		IStyle style = content.getStyle( );
		style.setProperty( IStyle.STYLE_MARGIN_BOTTOM, IStyle.NONE_VALUE );
		style.setProperty( IStyle.STYLE_BORDER_BOTTOM_STYLE, IStyle.NONE_VALUE );
		style.setProperty( IStyle.STYLE_PADDING_BOTTOM, IStyle.NONE_VALUE );
		return content;
	}

	protected boolean hasMasterPageChanged( IContent content )
	{
		IStyle style = content.getStyle( );
		String newMasterPage = style != null ? style.getMasterPage( ) : null;
		if ( newMasterPage == null )
		{
			return false;
		}
		if ( !newMasterPage.equalsIgnoreCase( masterPage ) )
		{
			//check if this master exist
			PageSetupDesign pageSetup = content.getReportContent( ).getDesign( ).getPageSetup( );
			if(pageSetup.getMasterPageCount( )>0)
			{
				MasterPageDesign masterPageDesign  = pageSetup.findMasterPage( newMasterPage );
				if(masterPageDesign!=null)
				{
					masterPage = newMasterPage;
					return true;
				}
			}
			
		}
		return false;
	}
	
	long getIndex( IContent content )
	{
		DocumentExtension docExt = (DocumentExtension) content
				.getExtension( IContent.DOCUMENT_EXTENSION );
		if ( docExt != null )
		{
			return docExt.getIndex( );
		}
		return -1;
	}
}